# -*- coding: utf-8 -*-
# Copyright (C) 2011-2016 Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>

"""Algorithms for factorizing argument dependent monomials."""

import numpy

from six import itervalues, iterkeys, iteritems
from six.moves import xrange as range
from itertools import chain

from ufl import as_ufl, conditional
from ufl.classes import Argument
from ufl.classes import Division
from ufl.classes import Product
from ufl.classes import Sum
from ufl.classes import Conditional
from ufl.classes import Zero
from ufl.algorithms import extract_type

from ffc.log import error

from ffc.uflacs.analysis.dependencies import compute_dependencies
from ffc.uflacs.analysis.modified_terminals import analyse_modified_terminal, strip_modified_terminal


def _build_arg_sets(V):
    "Build arg_sets = { argument number: set(j for j where V[j] is a modified Argument with this number) }"
    arg_sets = {}
    for i, v in enumerate(V):
        arg = strip_modified_terminal(v)
        if not isinstance(arg, Argument):
            continue
        num = arg.number()
        arg_set = arg_sets.get(num)
        if arg_set is None:
            arg_set = {}
            arg_sets[num] = arg_set
        arg_set[i] = v
    return arg_sets


def _build_argument_indices_from_arg_sets(V, arg_sets):
    "Build ordered list of indices to modified arguments."
    # Build set of all indices of V referring to modified arguments
    arg_indices = set()
    for js in itervalues(arg_sets):
        arg_indices.update(js)

    # Make a canonical ordering of vertex indices for modified arguments
    def arg_ordering_key(i):
        "Return a key for sorting argument vertex indices based on the properties of the modified terminal."
        mt = analyse_modified_terminal(arg_ordering_key.V[i])
        return mt.argument_ordering_key()
    arg_ordering_key.V = V
    ordered_arg_indices = sorted(arg_indices, key=arg_ordering_key)

    return ordered_arg_indices


def build_argument_indices(V):
    "Build ordered list of indices to modified arguments."
    arg_sets = _build_arg_sets(V)
    ordered_arg_indices = _build_argument_indices_from_arg_sets(V, arg_sets)
    return ordered_arg_indices


def build_argument_dependencies(dependencies, arg_indices):
    "Preliminary algorithm: build list of argument vertex indices each vertex (indirectly) depends on."
    n = len(dependencies)
    A = numpy.empty(n, dtype=object)
    for i, deps in enumerate(dependencies):
        argdeps = []
        for j in deps:
            if j in arg_indices:
                argdeps.append(j)
            else:
                argdeps.extend(A[j])
        A[i] = sorted(argdeps)
    return A


class Factors(object): # TODO: Refactor code in this file by using a class like this
    def __init__(self):
        self.FV = []
        self.e2fi = {}

    def add(self, expr):
        add_to_fv(expr, self.FV, self.e2fi)


def add_to_fv(expr, FV, e2fi):
    "Add expression expr to factor vector FV and expr->FVindex mapping e2fi."
    fi = e2fi.get(expr)
    if fi is None:
        fi = len(e2fi)
        FV.append(expr)
        e2fi[expr] = fi
    return fi


# Reuse these empty objects where appropriate to save memory
noargs = {}


def handle_sum(v, si, deps, SV_factors, FV, sv2fv, e2fi):
    if len(deps) != 2:
        error("Assuming binary sum here. This can be fixed if needed.")

    fac0 = SV_factors[deps[0]]
    fac1 = SV_factors[deps[1]]
    argkeys = set(fac0) | set(fac1)

    if argkeys:  # f*arg + g*arg = (f+g)*arg
        argkeys = sorted(argkeys)
        keylen = len(argkeys[0])
        factors = {}
        for argkey in argkeys:
            if len(argkey) != keylen:
                error("Expecting equal argument rank terms among summands.")

            fi0 = fac0.get(argkey)
            fi1 = fac1.get(argkey)
            if fi0 is None:
                fisum = fi1
            elif fi1 is None:
                fisum = fi0
            else:
                f0 = FV[fi0]
                f1 = FV[fi1]
                fisum = add_to_fv(f0 + f1, FV, e2fi)
            factors[argkey] = fisum

    else:  # non-arg + non-arg
        factors = noargs
        sv2fv[si] = add_to_fv(v, FV, e2fi)

    return factors


def handle_product(v, si, deps, SV_factors, FV, sv2fv, e2fi):
    if len(deps) != 2:
        error("Assuming binary product here. This can be fixed if needed.")
    fac0 = SV_factors[deps[0]]
    fac1 = SV_factors[deps[1]]

    if not fac0 and not fac1:  # non-arg * non-arg
        # Record non-argument product
        factors = noargs
        f0 = FV[sv2fv[deps[0]]]
        f1 = FV[sv2fv[deps[1]]]
        assert f1 * f0 == v
        sv2fv[si] = add_to_fv(v, FV, e2fi)
        assert FV[sv2fv[si]] == v

    elif not fac0:  # non-arg * arg
        # Record products of non-arg operand with each factor of arg-dependent operand
        f0 = FV[sv2fv[deps[0]]]
        factors = {}
        for k1 in sorted(fac1):
            fi1 = fac1[k1]
            factors[k1] = add_to_fv(f0 * FV[fi1], FV, e2fi)

    elif not fac1:  # arg * non-arg
        # Record products of non-arg operand with each factor of arg-dependent operand
        f1 = FV[sv2fv[deps[1]]]
        factors = {}
        for k0 in sorted(fac0):
            f0 = FV[fac0[k0]]
            factors[k0] = add_to_fv(f1 * f0, FV, e2fi)

    else:  # arg * arg
        # Record products of each factor of arg-dependent operand
        factors = {}
        for k0 in sorted(fac0):
            f0 = FV[fac0[k0]]
            for k1 in sorted(fac1):
                f1 = FV[fac1[k1]]
                argkey = tuple(sorted(k0 + k1))  # sort key for canonical representation
                factors[argkey] = add_to_fv(f0 * f1, FV, e2fi)

    return factors


def handle_division(v, si, deps, SV_factors, FV, sv2fv, e2fi):
    fac0 = SV_factors[deps[0]]
    fac1 = SV_factors[deps[1]]
    assert not fac1, "Cannot divide by arguments."

    if fac0:  # arg / non-arg
        # Record products of non-arg operand with each factor of arg-dependent operand
        f1 = FV[sv2fv[deps[1]]]
        factors = {}
        for k0 in sorted(fac0):
            f0 = FV[fac0[k0]]
            factors[k0] = add_to_fv(f0 / f1, FV, e2fi)

    else:  # non-arg / non-arg
        # Record non-argument subexpression
        factors = noargs
        sv2fv[si] = add_to_fv(v, FV, e2fi)

    return factors


def handle_conditional(v, si, deps, SV_factors, FV, sv2fv, e2fi):
    fac0 = SV_factors[deps[0]]
    fac1 = SV_factors[deps[1]]
    fac2 = SV_factors[deps[2]]
    assert not fac0, "Cannot have argument in condition."

    if not (fac1 or fac2):  # non-arg ? non-arg : non-arg
        # Record non-argument subexpression
        sv2fv[si] = add_to_fv(v, FV, e2fi)
        factors = noargs
    else:
        f0 = FV[sv2fv[deps[0]]]
        f1 = FV[sv2fv[deps[1]]]
        f2 = FV[sv2fv[deps[2]]]

        # Term conditional(c, argument, non-argument) is not legal unless non-argument is 0.0
        assert fac1 or isinstance(f1, Zero)
        assert fac2 or isinstance(f2, Zero)
        assert () not in fac1
        assert () not in fac2

        z = as_ufl(0.0)

        # In general, can decompose like this:
        #    conditional(c, sum_i fi*ui, sum_j fj*uj) -> sum_i conditional(c, fi, 0)*ui + sum_j conditional(c, 0, fj)*uj
        mas = sorted(set(fac1.keys()) | set(fac2.keys()))
        factors = {}
        for k in mas:
            fi1 = fac1.get(k)
            fi2 = fac2.get(k)
            f1 = z if fi1 is None else FV[fi1]
            f2 = z if fi2 is None else FV[fi2]
            factors[k] = add_to_fv(conditional(f0, f1, f2), FV, e2fi)

    return factors


def handle_operator(v, si, deps, SV_factors, FV, sv2fv, e2fi):
    # Error checking
    if any(SV_factors[d] for d in deps):
        error("Assuming that a {0} cannot be applied to arguments. If this is wrong please report a bug.".format(type(v)))
    # Record non-argument subexpression
    sv2fv[si] = add_to_fv(v, FV, e2fi)
    factors = noargs
    return factors


def compute_argument_factorization(SV, SV_deps, SV_targets, rank):
    """Factorizes a scalar expression graph w.r.t. scalar Argument
    components.

    The result is a triplet (AV, FV, IM):

      - The scalar argument component subgraph:

          AV[ai] = v

        with the property

          SV[arg_indices] == AV[:]

      - An expression graph vertex list with all non-argument factors:

          FV[fi] = f

        with the property that none of the expressions depend on Arguments.

      - A dict representation of the final integrand of rank r:

          IM = { (ai1_1, ..., ai1_r): fi1, (ai2_1, ..., ai2_r): fi2, }

        This mapping represents the factorization of SV[-1] w.r.t. Arguments s.t.:

          SV[-1] := sum(FV[fik] * product(AV[ai] for ai in aik) for aik, fik in IM.items())

        where := means equivalence in the mathematical sense,
        of course in a different technical representation.

    """
    # Extract argument component subgraph
    arg_indices = build_argument_indices(SV)
    #A = build_argument_dependencies(SV_deps, arg_indices)
    AV = [SV[si] for si in arg_indices]
    #av2sv = arg_indices
    sv2av = { si: ai for ai, si in enumerate(arg_indices) }
    assert all(AV[ai] == SV[si] for ai, si in enumerate(arg_indices))
    assert all(AV[ai] == SV[si] for si, ai in iteritems(sv2av))

    # Data structure for building non-argument factors
    FV = []
    e2fi = {}

    # Adding 0.0 as an expression to fix issue in conditional
    zero_index = add_to_fv(as_ufl(0.0), FV, e2fi)

    # Adding 1.0 as an expression allows avoiding special representation
    # of arguments when first visited by representing "v" as "1*v"
    one_index = add_to_fv(as_ufl(1.0), FV, e2fi)

    # Adding 2 as an expression fixes an issue with FV entries that change K*K -> K**2
    two_index = add_to_fv(as_ufl(2), FV, e2fi)

    # Intermediate factorization for each vertex in SV on the format
    # SV_factors[si] = None # if SV[si] does not depend on arguments
    # SV_factors[si] = { argkey: fi } # if SV[si] does depend on arguments, where:
    #   FV[fi] is the expression SV[si] with arguments factored out
    #   argkey is a tuple with indices into SV for each of the argument components SV[si] depends on
    # SV_factors[si] = { argkey1: fi1, argkey2: fi2, ... } # if SV[si] is a linear combination of multiple argkey configurations
    SV_factors = numpy.empty(len(SV), dtype=object)
    si2fi = numpy.zeros(len(SV), dtype=int)

    # Factorize each subexpression in order:
    for si, v in enumerate(SV):
        deps = SV_deps[si]

        # These handlers insert values in si2fi and SV_factors
        if not len(deps):
            if si in arg_indices:
                # v is a modified Argument
                factors = { (si,): one_index }
            else:
                # v is a modified non-Argument terminal
                si2fi[si] = add_to_fv(v, FV, e2fi)
                factors = noargs
        else:
            # These quantities could be better input args to handlers:
            #facs = [SV_factors[d] for d in deps]
            #fs = [FV[sv2fv[d]] for d in deps]
            if isinstance(v, Sum):
                handler = handle_sum
            elif isinstance(v, Product):
                handler = handle_product
            elif isinstance(v, Division):
                handler = handle_division
            elif isinstance(v, Conditional):
                handler = handle_conditional
            else:  # All other operators
                handler = handle_operator
            factors = handler(v, si, deps, SV_factors, FV, si2fi, e2fi)

        SV_factors[si] = factors

    assert not noargs, "This dict was not supposed to be filled with anything!"

    # Throw away superfluous items in array
    # FV = FV[:len(e2fi)]
    assert len(FV) == len(e2fi)

    # Get the factorizations of the target values
    IMs = []
    for si in SV_targets:
        if SV_factors[si] == {}:
            if rank == 0:
                # Functionals and expressions: store as no args * factor
                factors = { (): si2fi[si] }
            else:
                # Zero form of arity 1 or higher: make factors empty
                factors = {}
        else:
            # Forms of arity 1 or higher:
            # Map argkeys from indices into SV to indices into AV,
            # and resort keys for canonical representation
            factors = { tuple(sorted(sv2av[si] for si in argkey)): fi
                        for argkey, fi in SV_factors[si].items() }
        # Expecting all term keys to have length == rank
        # (this assumption will eventually have to change if we
        # implement joint bilinear+linear form factorization here)
        assert all(len(k) == rank for k in factors)
        IMs.append(factors)

    # Recompute dependencies in FV
    FV_deps = compute_dependencies(e2fi, FV)

    # Indices into FV that are needed for final result
    FV_targets = list(chain(sorted(IM.values())
                            for IM in IMs))

    return IMs, AV, FV, FV_deps, FV_targets


def rebuild_scalar_graph_from_factorization(AV, FV, IM):
    # TODO: What about multiple target_variables?

    # Build initial graph
    SV = []
    SV.extend(AV)
    SV.extend(FV)
    se2i = dict((s, i) for i, s in enumerate(SV))

    def add_vertex(h):
        # Avoid adding vertices twice
        i = se2i.get(h)
        if i is None:
            se2i[h] = len(SV)
            SV.append(h)

    # Add factorization monomials
    argkeys = sorted(iterkeys(IM))
    fs = []
    for argkey in argkeys:
        # Start with coefficients
        f = FV[IM[argkey]]
        # f = 1

        # Add binary products with each argument in order
        for argindex in argkey:
            f = f * AV[argindex]
            add_vertex(f)

        # Add product with coefficients last
        # f = f*FV[IM[argkey]]
        # add_vertex(f)

        # f is now the full monomial, store it as a term for sum below
        fs.append(f)

    # Add sum of factorization monomials
    g = 0
    for f in fs:
        g = g + f
        add_vertex(g)

    # Rebuild dependencies
    dependencies = compute_dependencies(se2i, SV)

    return SV, se2i, dependencies



# -*- coding: utf-8 -*-
"Code generator for tensor representation"

# Copyright (C) 2004-2013 Anders Logg
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Kristian B. Oelgaard, 2009-2010
# Modified by Marie Rognes, 2007
# Modified by Garth N. Wells, 2009
# Modified by Mehdi Nikbakht, 2010
# Modified by Martin Sandve Alnæs, 2013
#
# First added:  2004-11-03
# Last changed: 2013-02-10

# FFC modules
from ffc.log import error, warning
from ffc.cpp import format, remove_unused, count_ops

# FFC tensor representation modules
from ffc.tensor.monomialtransformation import MonomialIndex
from ffc.representationutils import initialize_integral_code


def generate_integral_code(ir, prefix, parameters):
    "Generate code for integral from intermediate representation."
    code = initialize_integral_code(ir, prefix, parameters)
    code["tabulate_tensor"] = _tabulate_tensor(ir, parameters)

    precision = ir["integrals_metadata"].get("precision")
    if precision is not None and precision != parameters["precision"]:
        warning("Ignoring precision in integral metadata compiled "
                "using tensor representation. Not implemented.")

    return code


def _tabulate_tensor(ir, parameters):
    "Generate code for tabulate_tensor."

    # Prefetch formats to speed up code generation
    comment = format["comment"]
    switch = format["switch"]

    # Set of used variables for Jacobian and geometry tensor
    j_set = set()
    g_set = set()

    # Extract data from intermediate representation
    AK = ir["AK"]
    integral_type = ir["integral_type"]
    tdim = ir["topological_dimension"]
    gdim = ir["geometric_dimension"]
    oriented = ir["needs_oriented"]
    num_facets = ir["num_facets"]

    # Check integral type and generate code
    if integral_type == "cell":

        # Generate code for one single tensor contraction
        t_code = _generate_tensor_contraction(AK, parameters, g_set)

        # Generate code for geometry tensors
        g_code = _generate_geometry_tensors(AK, j_set, g_set, tdim, gdim)

        # Generate code for basic geometric quantities
        j_code = ""
        j_code += format["compute_jacobian"](tdim, gdim)
        j_code += "\n"
        j_code += format["compute_jacobian_inverse"](tdim, gdim)
        if oriented:
            j_code += format["orientation"](tdim, gdim)
        j_code += "\n"
        j_code += format["scale factor snippet"]

    elif integral_type == "exterior_facet":

        # Generate code for num_facets tensor contractions
        cases = [None for i in range(num_facets)]
        for i in range(num_facets):
            cases[i] = _generate_tensor_contraction(AK[i], parameters, g_set)
        t_code = switch(format["facet"](None), cases)

        # Generate code for geometry tensors
        g_code = _generate_geometry_tensors(AK[0], j_set, g_set, tdim, gdim)

        # Generate code for Jacobian
        j_code = ""
        j_code += format["compute_jacobian"](tdim, gdim)
        j_code += "\n"
        j_code += format["compute_jacobian_inverse"](tdim, gdim)
        if oriented:
            j_code += format["orientation"](tdim, gdim)
        j_code += "\n"
        j_code += format["facet determinant"](tdim, gdim)

    elif integral_type == "interior_facet":

        # Generate code for num_facets x num_facets tensor contractions
        cases = [[None for j in range(num_facets)] for i in range(num_facets)]
        for i in range(num_facets):
            for j in range(num_facets):
                cases[i][j] = _generate_tensor_contraction(AK[i][j], parameters, g_set)
        t_code = switch(format["facet"]("+"), [switch(format["facet"]("-"), cases[i]) for i in range(len(cases))])

        # Generate code for geometry tensors
        g_code = _generate_geometry_tensors(AK[0][0], j_set, g_set, tdim, gdim)

        # Generate code for Jacobian
        j_code = ""
        for _r in ["+", "-"]:
            j_code += format["compute_jacobian"](tdim, gdim, r=_r)
            j_code += "\n"
            j_code += format["compute_jacobian_inverse"](tdim, gdim, r=_r)
            j_code += "\n"
            if oriented:
                j_code += format["orientation"](tdim, gdim, r=_r)
        j_code += format["facet determinant"](tdim, gdim, r="+")
        j_code += "\n"

    else:
        error("Unhandled integral type: " + str(integral_type))

    # Remove unused declarations from Jacobian code
    j_code = remove_unused(j_code, j_set)

    # Compute total number of operations
    j_ops, g_ops, t_ops = [count_ops(c) for c in (j_code, g_code, t_code)]
    total_ops = j_ops + g_ops + t_ops

    # Add generated code
    lines = []
    lines.append(comment("Number of operations (multiply-add pairs) for Jacobian data:      %d" % j_ops))
    lines.append(comment("Number of operations (multiply-add pairs) for geometry tensor:    %d" % g_ops))
    lines.append(comment("Number of operations (multiply-add pairs) for tensor contraction: %d" % t_ops))
    lines.append(comment("Total number of operations (multiply-add pairs):                  %d" % total_ops))
    lines.append("")
    lines.append(j_code)
    lines.append("")
    lines.append(comment("Compute geometry tensor"))
    lines.append(g_code)
    lines.append("")
    lines.append(comment("Compute element tensor"))
    lines.append(t_code)

    return "\n".join(lines)


def _generate_tensor_contraction(terms, parameters, g_set):
    """
    Generate code for computation of tensor contraction
    """
    return _generate_tensor_contraction_standard(terms, parameters, g_set)


def _generate_tensor_contraction_standard(terms, parameters, g_set):
    """
    Generate code for computation of tensor contraction using full
    tensor contraction.
    """

    # Prefetch formats to speed up code generation
    iadd = format["iadd"]
    assign = format["assign"]
    element_tensor = format["element tensor"]
    geometry_tensor = format["geometry tensor"]
    zero = format["float"](0)
    inner_product = format["inner product"]

    # True if we should add to element tensor (not used)
    incremental = False

    # Get machine precision
    epsilon = parameters["epsilon"]

    # Get list of primary indices (should be the same so pick first)
    A0, GK, optimized_contraction = terms[0]
    primary_indices = A0.primary_multi_index.indices

    # Generate code for geometry tensor entries
    gk_tensor = []
    for (j, (A0, GK, optimized_contraction)) in enumerate(terms):
        gk_tensor_j = []
        for a in A0.secondary_multi_index.indices:
            gk_tensor_j.append((geometry_tensor(j, a), a))
        gk_tensor.append((gk_tensor_j, j))

    # Generate code for computing the element tensor
    lines = []
    for (k, i) in enumerate(primary_indices):
        name = element_tensor(k)
        coefficients = []
        entries = []
        for (gka, j) in gk_tensor:
            (A0, GK, optimized_contraction) = terms[j]
            for (gk, a) in gka:
                a0 = A0.A0[tuple(i + a)]

                # Skip small values
                if abs(a0) < epsilon:
                    continue

                # Compute value
                coefficients.append(a0)
                entries.append(gk)

                # Remember that gk has been used
                g_set.add(gk)

        # Compute inner product
        value = inner_product(coefficients, entries)

        # Handle special case
        value = value or zero

        # Add value
        if incremental:
            lines.append(iadd(name, value))
        else:
            lines.append(assign(name, value))

    return "\n".join(lines)


def _generate_geometry_tensors(terms, j_set, g_set, tdim, gdim):
    "Generate code for computation of geometry tensors."

    # Prefetch formats to speed up code generation
    format_add = format["addition"]
    format_geometry_tensor = format["geometry tensor"]
    format_scale_factor = format["scale factor"]
    format_declaration = format["const float declaration"]

    # Iterate over all terms
    lines = []
    offset = 0
    det_used = False

    for (i, term) in enumerate(terms):

        # Get secondary indices
        A0, GK, optimized_contraction = term
        secondary_indices = GK.secondary_multi_index.indices

        # Hack to keep old code generation based on factorization of GK
        # in case we want to reimplement factorization
        GKs = [GK]

        # Iterate over secondary indices
        for a in secondary_indices:

            # Skip code generation if term is not used
            if not format["geometry tensor"](i, a) in g_set:
                continue

            # Compute factorized values
            values = [_generate_entry(GK, a, offset + j, j_set, tdim, gdim)
                      for (j, GK) in enumerate(GKs)]

            # Sum factorized values
            name = format_geometry_tensor(i, a)
            value = format_add(values)

            # Multiply with determinant factor
            dets = GK.determinants
            value = _multiply_value_by_det(value, dets, len(values) > 1, j_set)
            det_used = True

            # Add code
            lines.append(format_declaration(name, value))

        # Add to offset
        offset += len(GKs)

    # Add scale factor
    if det_used:
        j_set.add(format_scale_factor)  # meg says: If all values vanish, det is not used.

    return "\n".join(lines)


def _generate_entry(GK, a, i, j_set, tdim, gdim):
    "Generate code for the value of a GK entry."

    # Prefetch formats to speed up code generation
    grouping = format["grouping"]
    add = format["addition"]
    multiply = format["multiply"]

    # Compute product of factors outside sum
    factors = _extract_factors(GK, a, None, j_set, tdim, gdim,
                               MonomialIndex.SECONDARY)

    # Compute sum of products of factors inside sum
    terms = [multiply(_extract_factors(GK, a, b, j_set, tdim, gdim,
                                       MonomialIndex.EXTERNAL))
             for b in GK.external_multi_index.indices]

    # Compute product
    if factors:
        entry = multiply(factors + [grouping(add(terms))])
    else:
        entry = add(terms)

    return entry


def _multiply_value_by_det(value, dets, is_sum, j_set):
    "Generate code for multiplication of value by determinant(s)."

    # FIXME: MER: This is way complicated than it should be

    # Cell / exterior facets:
    d = []
    if all([det.restriction is None for det in dets]):
        total_power = sum(det.power for det in dets)
        if not total_power == 0:
            J = format["det(J)"](None)
            d += [format["power"](J, total_power)]
            j_set.add(J)
    # Interior facets
    else:
        for det in dets:
            if not det.power == 0:
                J = format["det(J)"](det.restriction)
                d += [format["power"](J, det.power)]
                j_set.add(J)

    if value == "1.0":
        v = []
    elif is_sum:
        v = [format["grouping"](value)]
    else:
        v = [value]

    return format["multiply"](d + [format["scale factor"]] + v)


def _extract_factors(GK, a, b, j_set, tdim, gdim, index_type):
    "Extract factors of given index type in GK entry."

    # Prefetch formats to speed up code generation
    coefficient = format["coefficient"]
    transform = format["transform"]

    # List of factors
    factors = []

    # Compute product of coefficients
    for c in GK.coefficients:
        if c.index.index_type == index_type:
            factors.append(coefficient(c.number, c.index(secondary=a)))

    # Compute product of transforms
    for t in GK.transforms:

        # Note non-trivial logic here
        if index_type == MonomialIndex.EXTERNAL:
            include_index = MonomialIndex.EXTERNAL in (t.index0.index_type, t.index1.index_type)
        else:
            include_index = not (MonomialIndex.EXTERNAL in (t.index0.index_type, t.index1.index_type))

        # Add factor
        if include_index:
            # FIXME: Dimensions of J and K are transposed, what is the right thing to fix this hack?
            if t.transform_type == "J":  # MonomialTransform.J:
                dim0, dim1 = gdim, tdim
            elif t.transform_type == "JINV":  # MonomialTransform.JINV:
                dim0, dim1 = tdim, gdim
            else:
                error("Unknown transform type, fix this hack.")

            factors.append(transform(t.transform_type,
                                     t.index0(secondary=a, external=b),
                                     t.index1(secondary=a, external=b),
                                     dim0, dim1,
                                     t.restriction))
            j_set.add(factors[-1])

    return factors



# -*- coding: utf-8 -*-
# Copyright (C) 2011-2016 Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.

"""Algorithms for working with computational graphs."""

import numpy

from six.moves import xrange as range

from ufl.classes import (GeometricQuantity, ConstantValue,
                         Argument, Coefficient,
                         Grad, Restricted, Indexed,
                         MathFunction)
from ufl.checks import is_cellwise_constant
from ffc.log import error

from ffc.uflacs.analysis.crsarray import CRSArray


def default_partition_seed(expr, rank):
    """
    Partition 0: Piecewise constant on each cell (including Real and DG0 coefficients)
    Partition 1: Depends on x
    Partition 2: Depends on x and coefficients
    Partitions [3,3+rank): depend on argument with count partition-3
    """
    # TODO: Use named constants for the partition numbers here

    modifiers = (Grad, Restricted, Indexed)  # FIXME: Add CellAvg, FacetAvg types here, others?
    if isinstance(expr, modifiers):
        return default_partition_seed(expr.ufl_operands[0], rank)

    elif isinstance(expr, Argument):
        ac = expr.number()
        assert 0 <= ac < rank
        poffset = 3
        p = poffset + ac
        return p

    elif isinstance(expr, Coefficient):
        if is_cellwise_constant(expr):  # This is crap, doesn't include grad modifier
            return 0
        else:
            return 2

    elif isinstance(expr, GeometricQuantity):
        if is_cellwise_constant(expr):  # This is crap, doesn't include grad modifier
            return 0
        else:
            return 1

    elif isinstance(expr, ConstantValue):
        return 0

    else:
        error("Don't know how to handle %s" % expr)


def mark_partitions(V, active, dependencies, rank,
                    partition_seed=default_partition_seed,
                    partition_combiner=max):
    """FIXME: Cover this with tests.

    Input:
    - V            - Array of expressions.
    - active       - Boolish array.
    - dependencies - CRSArray with V dependencies.
    - partition_seed - Policy for determining the partition of a terminalish.
    - partition_combiner - Policy for determinging the partition of an operator.

    Output:
    - partitions   - Array of partition int ids.
    """
    n = len(V)
    assert len(active) == n
    assert len(dependencies) == n
    partitions = numpy.zeros(n, dtype=int)
    for i, v in enumerate(V):
        deps = dependencies[i]
        if active[i]:
            if len(deps):
                p = partition_combiner([partitions[d] for d in deps])
            else:
                p = partition_seed(v, rank)
        else:
            p = -1
        partitions[i] = p
    return partitions


"""
def build_factorized_partitions():
    num_points = [3]

    # dofrange = (begin, end)
    # dofblock = ()  |  (dofrange0,)  |  (dofrange0, dofrange1)

    partitions = {}

    # partitions["piecewise"] = partition of expressions independent of quadrature and argument loops
    partitions["piecewise"] = []

    # partitions["varying"][np] = partition of expressions dependent on np quadrature but independent of argument loops
    partitions["varying"] = dict((np, []) for np in num_points)

    # partitions["argument"][np][iarg][dofrange] = partition depending on this dofrange of argument iarg
    partitions["argument"] = dict((np, [dict() for i in range(rank)]) for np in num_points)

    # partitions["integrand"][np][dofrange] = partition depending on this dofrange of argument iarg
    partitions["integrand"] = dict((np, dict()) for np in num_points)
"""

def compute_dependency_count(dependencies):
    """FIXME: Test"""
    n = len(dependencies)
    depcount = numpy.zeros(n, dtype=int)
    for i in range(n):
        for d in dependencies[i]:
            depcount[d] += 1
    return depcount


def invert_dependencies(dependencies, depcount):
    """FIXME: Test"""
    n = len(dependencies)
    m = sum(depcount)
    invdeps = [()] * n
    for i in range(n):
        for d in dependencies[i]:
            invdeps[d] = invdeps[d] + (i,)
    return CRSArray.from_rows(invdeps, n, m, int)


def default_cache_score_policy(vtype, ndeps, ninvdeps, partition):
    # Start at 1 and then multiply with various heuristic factors
    s = 1

    # Is the type particularly expensive to compute?
    expensive = (MathFunction,)
    if vtype in expensive:  # Could make a type-to-cost mapping, but this should do.
        s *= 20

    # More deps roughly correlates to more operations
    s *= ndeps

    # If it is reused several times let that count significantly
    s *= ninvdeps ** 3  # 1->1, 2->8, 3->27

    # Finally let partition count for something?
    # Or perhaps we need some more information, such as
    # when x from outer loop is used by y within inner loop.

    return s


def compute_cache_scores(V, active, dependencies, inverse_dependencies, partitions,
                         cache_score_policy=default_cache_score_policy):
    """FIXME: Cover with tests.

    TODO: Experiment with heuristics later when we have functional code generation.
    """
    n = len(V)
    score = numpy.zeros(n, dtype=int)
    for i, v in enumerate(V):
        if active[i]:
            deps = dependencies[i]
            ndeps = len(deps)
            invdeps = inverse_dependencies[i]
            ninvdeps = len(invdeps)
            p = partitions[i]
            s = cache_score_policy(type(v), ndeps, ninvdeps, p)
        else:
            s = -1
        score[i] = s
    return score


import heapq


def allocate_registers(active, partitions, targets,
                       scores, max_registers, score_threshold):
    """FIXME: Cover with tests.

    TODO: Allow reuse of registers, reducing memory usage.

    TODO: Probably want to sort within partitions.
    """
    # Check for consistent number of variables
    n = len(scores)
    assert n == len(active)
    assert n == len(partitions)
    num_targets = len(targets)

    # Analyse scores
    #min_score = min(scores)
    #max_score = max(scores)
    #mean_score = sum(scores) // n

    # Can allocate a number of registers up to given threshold
    num_to_allocate = max(num_targets,
                          min(max_registers, n) - num_targets)
    to_allocate = set()

    # For now, just using an arbitrary heuristic algorithm to select m largest scores
    queue = [(-scores[i], i) for i in range(n) if active[i]]
    heapq.heapify(queue)

    # Always allocate registers for all targets, for simplicity
    # in the rest of the code generation pipeline
    to_allocate.update(targets)

    # Allocate one register each for max_registers largest symbols
    for r in range(num_to_allocate):
        s, i = heapq.heappop(queue)
        if -s <= score_threshold:
            break
        if i in targets:
            continue
        to_allocate.add(i)

    registers_used = len(to_allocate)

    # Some consistency checks
    assert num_to_allocate <= max_registers
    assert registers_used <= num_to_allocate + len(targets)
    assert registers_used <= max(max_registers, len(targets))

    # Mark allocations
    allocations = numpy.zeros(n, dtype=int)
    allocations[:] = -1
    for r, i in enumerate(sorted(to_allocate)):
        allocations[i] = r

    # Possible data structures for improved register allocations
    # register_status = numpy.zeros(max_registers, dtype=int)

    # Stack/set of free registers (should wrap in stack abstraction):
    # free_registers = numpy.zeros(max_registers, dtype=int)
    # num_free_registers = max_registers
    # free_registers[:] = reversed(xrange(max_registers))

    return allocations



# -*- coding: utf-8 -*-
# Copyright (C) 2011-2016 Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>

"""FFC/UFC specific symbol naming."""

from ffc.log import error


# TODO: Get restriction postfix from somewhere central
def ufc_restriction_postfix(restriction):
    if restriction == "+":
        res = "_0"
    elif restriction == "-":
        res = "_1"
    else:
        res = ""
    return res


def format_mt_name(basename, mt):
    "Format variable name for modified terminal."
    access = str(basename)

    # Add averaged state to name
    if mt.averaged:
        avg = "_a{0}".format(mt.averaged)
        access += avg

    # Format restriction
    res = ufc_restriction_postfix(mt.restriction).replace("_", "_r")
    access += res

    # Format local derivatives
    assert not mt.global_derivatives
    if mt.local_derivatives:
        der = "_d{0}".format(''.join(map(str, mt.local_derivatives)))
        access += der

    # Add flattened component to name
    if mt.component:
        comp = "_c{0}".format(mt.flat_component)
        access += comp

    return access


class FFCBackendSymbols(object):
    """FFC specific symbol definitions. Provides non-ufl symbols."""
    def __init__(self, language, coefficient_numbering):
        self.L = language
        self.S = self.L.Symbol
        self.coefficient_numbering = coefficient_numbering

        # Used for padding variable names based on restriction
        self.restriction_postfix = { r: ufc_restriction_postfix(r)
                                     for r in ("+", "-", None) }


    def element_tensor(self):
        "Symbol for the element tensor itself."
        return self.S("A")


    def entity(self, entitytype, restriction):
        "Entity index for lookup in element tables."
        if entitytype == "cell":
            # Always 0 for cells (even with restriction)
            return self.L.LiteralInt(0)
        elif entitytype == "facet":
            return self.S("facet" + ufc_restriction_postfix(restriction))
        elif entitytype == "vertex":
            return self.S("vertex")
        else:
            error("Unknown entitytype {}".format(entitytype))


    def cell_orientation_argument(self, restriction):
        "Cell orientation argument in ufc. Not same as cell orientation in generated code."
        return self.S("cell_orientation" + ufc_restriction_postfix(restriction))


    def cell_orientation_internal(self, restriction):
        "Internal value for cell orientation in generated code."
        return self.S("co" + ufc_restriction_postfix(restriction))


    def argument_loop_index(self, iarg):
        "Loop index for argument #iarg."
        indices = ["i", "j", "k", "l"]
        return self.S(indices[iarg])


    def coefficient_dof_sum_index(self):
        """Index for loops over coefficient dofs, assumed to never be used in two nested loops."""
        return self.S("ic")


    def quadrature_loop_index(self):
        """Reusing a single index name for all quadrature loops,
        assumed not to be nested."""
        return self.S("iq")


    def num_custom_quadrature_points(self):
        "Number of quadrature points, argument to custom integrals."
        return self.S("num_quadrature_points")


    def custom_quadrature_weights(self):
        "Quadrature weights including cell measure scaling, argument to custom integrals."
        return self.S("quadrature_weights")


    def custom_quadrature_points(self):
        "Physical quadrature points, argument to custom integrals."
        return self.S("quadrature_points")


    def custom_weights_table(self):
        "Table for chunk of custom quadrature weights (including cell measure scaling)."
        return self.S("weights_chunk")


    def custom_points_table(self):
        "Table for chunk of custom quadrature points (physical coordinates)."
        return self.S("points_chunk")


    def weights_table(self, num_points):
        "Table of quadrature weights."
        return self.S("weights%d" % (num_points,))


    def points_table(self, num_points):
        "Table of quadrature points (points on the reference integration entity)."
        return self.S("points%d" % (num_points,))


    def x_component(self, mt):
        "Physical coordinate component."
        return self.S(format_mt_name("x", mt))


    def X_component(self, mt):
        "Reference coordinate component."
        return self.S(format_mt_name("X", mt))


    def J_component(self, mt):
        "Jacobian component."
        # FIXME: Add domain number!
        return self.S(format_mt_name("J", mt))


    def domain_dof_access(self, dof, component, gdim, num_scalar_dofs,
                          restriction, interleaved_components):
        # FIXME: Add domain number or offset!
        vc = self.S("coordinate_dofs" + ufc_restriction_postfix(restriction))
        if interleaved_components:
            return vc[gdim*dof + component]
        else:
            return vc[num_scalar_dofs*component + dof]


    def domain_dofs_access(self, gdim, num_scalar_dofs, restriction,
                           interleaved_components):
        # FIXME: Add domain number or offset!
        return [self.domain_dof_access(dof, component, gdim, num_scalar_dofs,
                                       restriction, interleaved_components)
                for component in range(gdim)
                for dof in range(num_scalar_dofs)]


    def coefficient_dof_access(self, coefficient, dof_number):
        # TODO: Add domain number?
        c = self.coefficient_numbering[coefficient]
        w = self.S("w")
        return w[c, dof_number]


    def coefficient_value(self, mt):
        "Symbol for variable holding value or derivative component of coefficient."
        c = self.coefficient_numbering[mt.terminal]
        return self.S(format_mt_name("w%d" % (c,), mt))


    def element_table(self, tabledata, entitytype, restriction):
        if tabledata.is_uniform:
            entity = 0
        else:
            entity = self.entity(entitytype, restriction)

        if tabledata.is_piecewise:
            iq = 0
        else:
            iq = self.quadrature_loop_index()

        # Return direct access to element table
        return self.S(tabledata.name)[entity][iq]



# -*- coding: utf-8 -*-
# Copyright (C) 2011 Marie E. Rognes
#
# This file is part of DOLFIN.
#
# DOLFIN is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DOLFIN is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with DOLFIN. If not, see <http://www.gnu.org/licenses/>.
#
# Based on original implementation by Martin Alnes and Anders Logg
#
# Modified by Anders Logg 2015
#
# Last changed: 2015-11-05

from . import includes as incl
from .functionspace import *
from .form import generate_form
from .capsules import UFCElementNames

__all__ = ["generate_dolfin_code"]

# NB: generate_dolfin_namespace(...) assumes that if a coefficient has
# the same name in multiple forms, it is indeed the same coefficient:
parameters = {"use_common_coefficient_names": True}


def generate_dolfin_code(prefix, header, forms,
                         common_function_space=False, add_guards=False,
                         error_control=False):
    """Generate complete dolfin wrapper code with given generated names.

    @param prefix:
        String, prefix for all form names.
    @param header:
        Code that will be inserted at the top of the file.
    @param forms:
        List of UFCFormNames instances or single UFCElementNames.
    @param common_function_space:
        True if common function space, otherwise False
    @param add_guards:
        True iff guards (ifdefs) should be added
    @param error_control:
        True iff adaptivity typedefs (ifdefs) should be added
    """

    # Generate dolfin namespace
    namespace = generate_dolfin_namespace(prefix, forms, common_function_space,
                                          error_control)

    # Collect pieces of code
    code = [incl.dolfin_tag, header, incl.stl_includes, incl.dolfin_includes,
            namespace]

    # Add ifdefs/endifs if specified
    if add_guards:
        guard_name = ("%s_h" % prefix).upper()
        preguard = "#ifndef %s\n#define %s\n" % (guard_name, guard_name)
        postguard = "\n#endif\n\n"
        code = [preguard] + code + [postguard]

    # Return code
    return "\n".join(code)


def generate_dolfin_namespace(prefix, forms, common_function_space=False,
                              error_control=False):

    # Allow forms to represent a single space, and treat separately
    if isinstance(forms, UFCElementNames):
        return generate_single_function_space(prefix, forms)

    # Extract (common) coefficient spaces
    assert(parameters["use_common_coefficient_names"])
    spaces = extract_coefficient_spaces(forms)

    # Generate code for common coefficient spaces
    code = [apply_function_space_template(*space) for space in spaces]

    # Generate code for forms
    code += [generate_form(form, "Form_%s" % form.name, error_control) for form in forms]

    # Generate namespace typedefs (Bilinear/Linear & Test/Trial/Function)
    code += [generate_namespace_typedefs(forms, common_function_space,
                                         error_control)]

    # Wrap code in namespace block
    code = "\nnamespace %s\n{\n\n%s\n}" % (prefix, "\n".join(code))

    # Return code
    return code


def generate_single_function_space(prefix, space):
    code = apply_function_space_template("FunctionSpace",
                                         space.ufc_finite_element_classnames[0],
                                         space.ufc_dofmap_classnames[0])
    code = "\nnamespace %s\n{\n\n%s\n}" % (prefix, code)
    return code


def generate_namespace_typedefs(forms, common_function_space, error_control):

    # Generate typedefs as (fro, to) pairs of strings
    pairs = []

    # Add typedef for Functional/LinearForm/BilinearForm if only one
    # is present of each
    aliases = ["Functional", "LinearForm", "BilinearForm"]
    extra_aliases = {"LinearForm": "ResidualForm",
                     "BilinearForm": "JacobianForm"}
    for rank in sorted(range(len(aliases)), reverse=True):
        forms_of_rank = [form for form in forms if form.rank == rank]
        if len(forms_of_rank) == 1:
            pairs += [("Form_%s" % forms_of_rank[0].name, aliases[rank])]
            if not error_control:  # FIXME: Issue #91
                pairs += [("MultiMeshForm_%s" % forms_of_rank[0].name,
                           "MultiMesh" + aliases[rank])]
            if aliases[rank] in extra_aliases:
                extra_alias = extra_aliases[aliases[rank]]
                pairs += [("Form_%s" % forms_of_rank[0].name, extra_alias)]
                if not error_control:  # FIXME: Issue #91
                    pairs += [("MultiMeshForm_%s" % forms_of_rank[0].name,
                               "MultiMesh" + extra_alias)]

    # Keepin' it simple: Add typedef for FunctionSpace if term applies
    if common_function_space:
        for i, form in enumerate(forms):
            if form.rank:
                pairs += [("Form_%s::TestSpace" % form.name, "FunctionSpace")]
                if not error_control:  # FIXME: Issue #91
                    pairs += [("Form_%s::MultiMeshTestSpace" % form.name,
                               "MultiMeshFunctionSpace")]
                break

    # Add specialized typedefs when adding error control wrapppers
    if error_control:
        pairs += error_control_pairs(forms)

    # Combine data to typedef code
    typedefs = "\n".join("typedef %s %s;" % (to, fro) for (to, fro) in pairs)

    # Return typedefs or ""
    if not typedefs:
        return ""
    return "// Class typedefs\n" + typedefs + "\n"


def error_control_pairs(forms):
    assert (len(forms) == 11), "Expecting 11 error control forms"

    return [("Form_%s" % forms[8].name, "BilinearForm"),
            ("Form_%s" % forms[9].name, "LinearForm"),
            ("Form_%s" % forms[10].name, "GoalFunctional")]



# -*- coding: utf-8 -*-
"Extraction of monomial representations of UFL forms."

# Copyright (C) 2008-2013 Anders Logg
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Martin Sandve Alnæs, 2008, 2013
# Modified by Kristian B. Oelgaard
#
# First added:  2008-08-01
# Last changed: 2013-01-08

# UFL modules
from ufl.classes import Form, Argument, Coefficient, ScalarValue, IntValue
from ufl.algorithms import purge_list_tensors, apply_transformer, ReuseTransformer

# FFC modules
from ffc.log import info, debug, ffc_assert

# Cache for computed integrand representations
#_cache = {}

def extract_monomial_form(integrands, function_replace_map):
    """
    Extract monomial representation of form (if possible). When
    successful, the form is represented as a sum of products of scalar
    components of basis functions or derivatives of basis functions.
    If unsuccessful, MonomialException is raised.
    """

    info("Extracting monomial form representation from UFL form")

    # Iterate over all integrals
    monomial_form = MonomialForm()
    for integrand in integrands:
        # Extract monomial representation if possible
        monomial_integrand = extract_monomial_integrand(integrand, function_replace_map)
        monomial_form.append(monomial_integrand)

    return monomial_form

def extract_monomial_integrand(integrand, function_replace_map):
    "Extract monomial integrand (if possible)."

    # Check cache
    #if integrand in _cache:
    #    debug("Reusing monomial integrand from cache")
    #    return _cache[integrand]

    # Purge list tensors
    integrand = purge_list_tensors(integrand)

    # Apply monomial transformer
    monomial_integrand = apply_transformer(integrand, MonomialTransformer(function_replace_map))

    # Store in cache
    #_cache[integrand] = monomial_integrand

    return monomial_integrand

class MonomialException(Exception):
    "Exception raised when monomial extraction fails."
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)

class MonomialFactor:
    """
    This class represents a monomial factor, that is, a derivative of
    a scalar component of a basis function.
    """

    def __init__(self, arg=None):
        if isinstance(arg, MonomialFactor):
            self.function = arg.function
            self.components = arg.components
            self.derivatives = arg.derivatives
            self.restriction = arg.restriction
        elif isinstance(arg, (Argument, Coefficient)):
            self.function = arg
            self.components = []
            self.derivatives = []
            self.restriction = None
            if isinstance(arg, Argument) and arg.part() is not None: # Not supported (yet?)
                raise MonomialException("Unable to create monomial from expression: " + str(arg))
        elif arg is None:
            self.function = None
            self.components = []
            self.derivatives = []
            self.restriction = None
        else:
            raise MonomialException("Unable to create monomial from expression: " + str(arg))

    def element(self):
        return self.function.ufl_element()

    def index(self):
        if isinstance(self.function, Coefficient):
            return self.function.count()
        else:
            return self.function.number()

    def apply_derivative(self, indices):
        self.derivatives += indices

    def apply_restriction(self, restriction):
        self.restriction = restriction

    def replace_indices(self, old_indices, new_indices):
        if old_indices is None:
            self.components = new_indices
        else:
            _replace_indices(self.components, old_indices, new_indices)
            _replace_indices(self.derivatives, old_indices, new_indices)

    def __str__(self):
        if len(self.components) == 0:
            c = ""
        else:
            c = "[%s]" % ", ".join(str(c) for c in self.components)
        if len(self.derivatives) == 0:
            d0 = ""
            d1 = ""
        else:
            d0 = "(" + " ".join("d/dx_%s" % str(d) for d in self.derivatives) + " "
            d1 = ")"
        if self.restriction is None:
            r = ""
        else:
            r = "(%s)" % str(self.restriction)
        return d0 + str(self.function) + r + c + d1

class Monomial:
    "This class represents a product of monomial factors."

    def __init__(self, arg=None):
        if isinstance(arg, Monomial):
            self.float_value = arg.float_value
            self.factors = [MonomialFactor(v) for v in arg.factors]
            self.index_slots = arg.index_slots
        elif isinstance(arg, (MonomialFactor, Argument, Coefficient)):
            self.float_value = 1.0
            self.factors = [MonomialFactor(arg)]
            self.index_slots = None
        elif isinstance(arg, ScalarValue):
            self.float_value = float(arg)
            self.factors = []
            self.index_slots = None
        elif arg is None:
            self.float_value = 1.0
            self.factors = []
            self.index_slots = None
        else:
            raise MonomialException("Unable to create monomial from expression: " + str(arg))

    def apply_derivative(self, indices):
        if len(self.factors) != 1:
            raise MonomialException("Expecting a single factor.")
        self.factors[0].apply_derivative(indices)

    def apply_tensor(self, indices):
        if self.index_slots is not None:
            raise MonomialException("Expecting scalar-valued expression.")
        self.index_slots = indices

    def apply_indices(self, indices):
        for v in self.factors:
            v.replace_indices(self.index_slots, indices)
        self.index_slots = None

    def apply_restriction(self, restriction):
        for v in self.factors:
            v.apply_restriction(restriction)

    def __mul__(self, other):
        m = Monomial()
        m.float_value = self.float_value * other.float_value
        m.factors = self.factors + other.factors
        return m

    def __str__(self):
        if self.float_value == 1.0:
            float_value = ""
        else:
            float_value = "%g * " % self.float_value
        return float_value + " * ".join(str(v) for v in self.factors)

class MonomialSum:
    "This class represents a sum of monomials."

    def __init__(self, arg=None):
        if isinstance(arg, MonomialSum):
            self.monomials = [Monomial(m) for m in arg.monomials]
        elif arg is None:
            self.monomials = []
        else:
            self.monomials = [Monomial(arg)]

    def apply_derivative(self, indices):
        for m in self.monomials:
            m.apply_derivative(indices)

    def apply_tensor(self, indices):
        for m in self.monomials:
            m.apply_tensor(indices)

    def apply_indices(self, indices):
        for m in self.monomials:
            m.apply_indices(indices)

    def apply_restriction(self, restriction):
        for m in self.monomials:
            m.apply_restriction(restriction)

    def __add__(self, other):
        m0 = [Monomial(m) for m in self.monomials]
        m1 = [Monomial(m) for m in other.monomials]
        sum = MonomialSum()
        sum.monomials = m0 + m1
        return sum

    def __mul__(self, other):
        sum = MonomialSum()
        for m0 in self.monomials:
            for m1 in other.monomials:
                sum.monomials.append(m0 * m1)
        return sum

    def __str__(self):
        return " + ".join(str(m) for m in self.monomials)

class MonomialForm:
    """
    This class represents a monomial form, that is, a sum of
    integrals, each represented as a MonomialSum.
    """

    def __init__(self):
        self.integrands = []

    def append(self, integrand):
        self.integrands.append(integrand)

    def __len__(self):
        return len(self.integrands)

    def __getitem__(self, i):
        return self.integrands[i]

    def __iter__(self):
        return iter(self.integrands)

    def __str__(self):
        if len(self.integrands) == 0:
            return "<Empty form>"
        s  = "Monomial form of %d integral(s)\n" % len(self.integrands)
        s += len(s) * "-" + "\n"
        for integrand in self.integrands:
            s += "Integrand: " + str(integrand) + "\n"
        return s

class MonomialTransformer(ReuseTransformer):
    """
    This class defines the transformation rules for extraction of a
    monomial form represented as a MonomialSum from a UFL integral.
    """

    def __init__(self, function_replace_map=None):
        ReuseTransformer.__init__(self)
        self._function_replace_map = function_replace_map or {}

    def expr(self, o, *ops):
        raise MonomialException("No handler defined for expression %s." % o._ufl_class_.__name__)

    def terminal(self, o):
        raise MonomialException("No handler defined for terminal %s." % o._ufl_class_.__name__)

    def variable(self, o):
        return self.visit(o.expression())

    #--- Operator handles ---

    def division(self, o):

        # Handle division by scalars as multiplication by inverse
        denominator = o.ufl_operands[1]
        if not isinstance(denominator, ScalarValue):
            raise MonomialException("No handler defined for expression %s."
                                    % o._ufl_class_.__name__)
        inverse = self.scalar_value(ScalarValue(1.0/denominator.value()))
        numerator = self.visit(o.ufl_operands[0])

        return self.product(o, inverse, numerator)

    def sum(self, o, s0, s1):
        s = s0 + s1
        return s

    def product(self, o, s0, s1):
        s = s0 * s1
        return s

    def index_sum(self, o, s, index):
        return s

    def indexed(self, o, s, indices):
        s = MonomialSum(s)
        s.apply_indices(indices)
        return s

    def component_tensor(self, o, s, indices):
        s = MonomialSum(s)
        s.apply_tensor(indices)
        return s

    def grad(self, o, s):
        # The representation
        #   o = Grad(s)
        # is equivalent to
        #   o = as_tensor(s[ii].dx(i), ii+(i,))
        # with ii a tuple of free indices and i a single free index.

        # Make some unique utility indices
        from ufl import indices
        on = len(o.ufl_shape)
        ind = list(indices(on))

        # Get underlying expression to take gradient of
        f, = o.ufl_operands
        fn = len(f.ufl_shape)

        ffc_assert(on == fn + 1, "Assuming grad always adds one axis.")

        s = MonomialSum(s)
        s.apply_indices(list(ind[:-1]))

        s = MonomialSum(s)
        s.apply_derivative([ind[-1]])

        s = MonomialSum(s)
        s.apply_tensor(ind)
        return s

    def positive_restricted(self, o, s):
        s.apply_restriction("+")
        return s

    def negative_restricted(self, o, s):
        s.apply_restriction("-")
        return s

    def power(self, o, s, ignored_exponent_expressed_as_sum):
        (expr, exponent) = o.ufl_operands
        if not isinstance(exponent, IntValue):
            raise MonomialException("Cannot handle non-integer exponents.")
        e = int(exponent)
        if e < 0:
            raise MonomialException("Cannot handle negative exponents.")
        p = MonomialSum(Monomial())
        for i in range(e):
            p = p * s
        return p

    #--- Terminal handlers ---

    def multi_index(self, multi_index):
        indices = [index for index in multi_index]
        return indices

    def argument(self, v):
        s = MonomialSum(self._function_replace_map.get(v,v))
        return s

    def coefficient(self, v):
        s = MonomialSum(self._function_replace_map.get(v,v))
        return s

    def scalar_value(self, x):
        s = MonomialSum(x)
        return s

def _replace_indices(indices, old_indices, new_indices):
    "Handle replacement of subsets of multi indices."

    # Old and new indices must match
    if len(old_indices) != len(new_indices):
        raise MonomialException("Unable to replace indices, mismatching index dimensions.")

    # Build index map
    index_map = {}
    for (i, index) in enumerate(old_indices):
        index_map[index] = new_indices[i]

    # Check all indices and replace
    for (i, index) in enumerate(indices):
        if index in old_indices:
            indices[i] = index_map[index]



# -*- coding: utf-8 -*-
"This file implements a class to represent a symbol."

# Copyright (C) 2009-2011 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# First added:  2009-07-12
# Last changed: 2011-06-28

# FFC modules.
from ffc.log import error

# FFC quadrature modules.
from .symbolics import type_to_string
from .symbolics import create_float
from .symbolics import create_product
from .symbolics import create_sum
from .symbolics import create_fraction
from .expr import Expr


class Symbol(Expr):
    __slots__ = ("v", "base_expr", "base_op", "exp", "cond")

    def __init__(self, variable, symbol_type, base_expr=None, base_op=0):
        """Initialise a Symbols object, it derives from Expr and contains
        the additional variables:

        v         - string, variable name
        base_expr - Other expression type like 'x*y + z'
        base_op   - number of operations for the symbol itself if it's a math
                    operation like std::cos(.) -> base_op = 1.
        NOTE: self._prec = 1."""

        # Dummy value, a symbol is always one.
        self.val = 1.0

        # Initialise variable, type and class.
        self.v = variable
        self.t = symbol_type
        self._prec = 1

        # Needed for symbols like std::cos(x*y + z),
        # where base_expr = x*y + z.
        # ops = base_expr.ops() + base_ops = 2 + 1 = 3
        self.base_expr = base_expr
        self.base_op = base_op

        # If type of the base_expr is lower than the given symbol_type change type.
        # TODO: Should we raise an error here? Or simply require that one
        # initalise the symbol by Symbol('std::cos(x*y)', (x*y).t, x*y, 1).
        if base_expr and base_expr.t < self.t:
            self.t = base_expr.t

        # Compute the representation now, such that we can use it directly
        # in the __eq__ and __ne__ methods (improves performance a bit, but
        # only when objects are cached).
        if self.base_expr:  # and self.exp is None:
            self._repr = "Symbol('%s', %s, %s, %d)" % (self.v, type_to_string[self.t],
                                                       self.base_expr._repr, self.base_op)
        else:
            self._repr = "Symbol('%s', %s)" % (self.v, type_to_string[self.t])

        # Use repr as hash value.
        self._hash = hash(self._repr)

    # Print functions.
    def __str__(self):
        "Simple string representation which will appear in the generated code."
#        print "sym str: ", self.v
        return self.v

    # Binary operators.
    def __add__(self, other):
        "Addition by other objects."
        # NOTE: We expect expanded objects
        # symbols, if other is a product, try to let product handle the addition.
        # Returns x + x -> 2*x, x + 2*x -> 3*x.
        if self._repr == other._repr:
            return create_product([create_float(2), self])
        elif other._prec == 2:  # prod
            return other.__add__(self)
        return create_sum([self, other])

    def __sub__(self, other):
        "Subtract other objects."
        # NOTE: We expect expanded objects
        # symbols, if other is a product, try to let product handle the addition.
        if self._repr == other._repr:
            return create_float(0)
        elif other._prec == 2:  # prod
            if other.get_vrs() == (self,):
                return create_product([create_float(1.0 - other.val), self]).expand()
        return create_sum([self, create_product([create_float(-1), other])])

    def __mul__(self, other):
        "Multiplication by other objects."
        # NOTE: We assume expanded objects.
        # If product will be zero.
        if self.val == 0.0 or other.val == 0.0:
            return create_float(0)

        # If other is Sum or Fraction let them handle the multiply.
        if other._prec in (3, 4):  # sum or frac
            return other.__mul__(self)

        # If other is a float or symbol, create simple product.
        if other._prec in (0, 1):  # float or sym
            return create_product([self, other])

        # Else add variables from product.
        return create_product([self] + other.vrs)

    def __truediv__(self, other):
        "Division by other objects."
        # NOTE: We assume expanded objects.
        # If division is illegal (this should definitely not happen).
        if other.val == 0.0:
            error("Division by zero.")

        # Return 1 if the two symbols are equal.
        if self._repr == other._repr:
            return create_float(1)

        # If other is a Sum we can only return a fraction.
        # TODO: Refine this later such that x / (x + x*y) -> 1 / (1 + y)?
        if other._prec == 3:  # sum
            return create_fraction(self, other)

        # Handle division by FloatValue, Symbol, Product and Fraction.
        # Create numerator and list for denominator.
        num = [self]
        denom = []

        # Add floatvalue, symbol and products to the list of denominators.
        if other._prec in (0, 1):  # float or sym
            denom = [other]
        elif other._prec == 2:  # prod
            # Need copies, so can't just do denom = other.vrs.
            denom += other.vrs
        # fraction.
        else:
            # TODO: Should we also support division by fraction for generality?
            # It should not be needed by this module.
            error("Did not expected to divide by fraction.")

        # Remove one instance of self in numerator and denominator if
        # present in denominator i.e., x/(x*y) --> 1/y.
        if self in denom:
            denom.remove(self)
            num.remove(self)

        # Loop entries in denominator and move float value to numerator.
        for d in denom:
            # Add the inverse of a float to the numerator, remove it from
            # the denominator and continue.
            if d._prec == 0:  # float
                num.append(create_float(1.0 / other.val))
                denom.remove(d)
                continue

        # Create appropriate return value depending on remaining data.
        # Can only be for x / (2*y*z) -> 0.5*x / (y*z).
        if len(num) > 1:
            num = create_product(num)
        # x / (y*z) -> x/(y*z),
        elif num:
            num = num[0]
        # else x / (x*y) -> 1/y.
        else:
            num = create_float(1)

        # If we have a long denominator, create product and fraction.
        if len(denom) > 1:
            return create_fraction(num, create_product(denom))
        # If we do have a denominator, but only one variable don't create a
        # product, just return a fraction using the variable as denominator.
        elif denom:
            return create_fraction(num, denom[0])
        # If we don't have any donominator left, return the numerator.
        # x / 2.0 -> 0.5*x.
        return num.expand()

    __div__ = __truediv__

    # Public functions.
    def get_unique_vars(self, var_type):
        "Get unique variables (Symbols) as a set."
        # Return self if type matches, also return base expression variables.
        s = set()
        if self.t == var_type:
            s.add(self)
        if self.base_expr:
            s.update(self.base_expr.get_unique_vars(var_type))
        return s

    def get_var_occurrences(self):
        """Determine the number of times all variables occurs in the expression.
        Returns a dictionary of variables and the number of times they occur."""
        # There is only one symbol.
        return {self: 1}

    def ops(self):
        "Returning the number of floating point operation for symbol."
        # Get base ops, typically 1 for sin() and then add the operations
        # for the base (sin(2*x + 1)) --> 2 + 1.
        if self.base_expr:
            return self.base_op + self.base_expr.ops()
        return self.base_op



# -*- coding: utf-8 -*-
# Copyright (C) 2011-2016 Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.

"""Tools for analysing dependencies within expression graphs."""

import numpy

from six.moves import xrange as range

from ffc.uflacs.analysis.crsarray import CRSArray, sufficient_int


def compute_dependencies(e2i, V, ignore_terminal_modifiers=True):
    # Use numpy int type sufficient to hold num_rows
    num_rows = len(V)
    itype = sufficient_int(num_rows)

    # Preallocate CRSArray matrix of sufficient capacity
    num_nonzeros = sum(len(v.ufl_operands) for v in V)
    dependencies = CRSArray(num_rows, num_nonzeros, itype)
    for v in V:
        if v._ufl_is_terminal_ or (ignore_terminal_modifiers and v._ufl_is_terminal_modifier_):
            dependencies.push_row(())
        else:
            dependencies.push_row([e2i[o] for o in v.ufl_operands])

    return dependencies


def mark_active(dependencies, targets):
    """Return an array marking the recursive dependencies of targets.

    Input:
    - dependencies - CRSArray of ints, a mapping from a symbol to the symbols of its dependencies.
    - targets      - Sequence of symbols to mark the dependencies of.

    Output:
    - active   - Truth value for each symbol.
    - num_used - Number of true values in active array.
    """
    n = len(dependencies)

    # Initial state where nothing is marked as used
    active = numpy.zeros(n, dtype=numpy.int8)
    num_used = 0

    # Seed with initially used symbols
    active[targets] = 1

    # Mark dependencies by looping backwards through symbols array
    for s in range(n - 1, -1, -1):
        if active[s]:
            num_used += 1
            active[dependencies[s]] = 1

    # Return array marking which symbols are used and the number of positives
    return active, num_used


def mark_image(inverse_dependencies, sources):
    """Return an array marking the set of symbols dependent on the sources.

    Input:
    - dependencies - CRSArray of ints, a mapping from a symbol to the symbols of its dependencies.
    - sources      - Sequence of symbols to mark the dependants of.

    Output:
    - image    - Truth value for each symbol.
    - num_used - Number of true values in active array.
    """
    n = len(inverse_dependencies)

    # Initial state where nothing is marked as used
    image = numpy.zeros(n, dtype=numpy.int8)
    num_used = 0

    # Seed with initially used symbols
    image[sources] = 1

    # Mark dependencies by looping forwards through symbols array
    for s in range(n):
        if image[s]:
            num_used += 1
            image[inverse_dependencies[s]] = 1

    # Return array marking which symbols are used and the number of positives
    return image, num_used



# -*- coding: utf-8 -*-

# Copyright (C) 2007-2016 Anders Logg, Martin Alnaes, Kristian B. Oelgaard,
# and others
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
Compiler stage 1: Analysis
--------------------------

This module implements the analysis/preprocessing of variational
forms, including automatic selection of elements, degrees and
form representation type.
"""

import numpy
from six import string_types

import os
import copy
from itertools import chain

# UFL modules
from ufl.classes import Form, CellVolume, FacetArea
from ufl.integral import Integral
from ufl.finiteelement import MixedElement, EnrichedElement, VectorElement
from ufl.algorithms import sort_elements
from ufl.algorithms import compute_form_data
from ufl.algorithms.analysis import extract_sub_elements
from ufl import custom_integral_types

# FFC modules
from ffc.log import info, begin, end, warning, debug, error, ffc_assert, warning_blue
from ffc.utils import all_equal
from ffc.tensor import estimate_cost
from ffc.cpp import default_precision


def analyze_forms(forms, parameters):
    """
    Analyze form(s), returning

       form_datas      - a tuple of form_data objects
       unique_elements - a tuple of unique elements across all forms
       element_numbers - a mapping to unique numbers for all elements
    """
    return analyze_ufl_objects(forms, "form", parameters)


def analyze_elements(elements, parameters):
    return analyze_ufl_objects(elements, "element", parameters)


def analyze_coordinate_mappings(coordinate_elements, parameters):
    return analyze_ufl_objects(coordinate_elements, "coordinate_mapping", parameters)


def analyze_ufl_objects(ufl_objects, kind, parameters):
    """
    Analyze ufl object(s), either forms, elements, or coordinate mappings, returning:

       form_datas      - a tuple of form_data objects
       unique_elements - a tuple of unique elements across all forms
       element_numbers - a mapping to unique numbers for all elements

    """
    begin("Compiler stage 1: Analyzing %s(s)" % (kind,))

    form_datas = ()
    unique_elements = set()
    unique_coordinate_elements = set()

    if kind == "form":
        forms = ufl_objects

        # Analyze forms
        form_datas = tuple(_analyze_form(form, parameters)
                           for form in forms)

        # Extract unique elements accross all forms
        for form_data in form_datas:
            unique_elements.update(form_data.unique_sub_elements)

        # Extract coordinate elements across all forms
        for form_data in form_datas:
            unique_coordinate_elements.update(form_data.coordinate_elements)

    elif kind == "element":
        elements = ufl_objects

        # Extract unique (sub)elements
        unique_elements.update(extract_sub_elements(elements))

    elif kind == "coordinate_mapping":
        meshes = ufl_objects

        # Extract unique (sub)elements
        unique_coordinate_elements = [mesh.ufl_coordinate_element() for mesh in meshes]

    # Make sure coordinate elements and their subelements are included
    unique_elements.update(extract_sub_elements(unique_coordinate_elements))

    # Sort elements
    unique_elements = sort_elements(unique_elements)

    # Check for schemes for QuadratureElements
    for element in unique_elements:
        if element.family() == "Quadrature":
            qs = element.quadrature_scheme()
            if qs is None:
                error("Missing quad_scheme in quadrature element.")

    # Compute element numbers
    element_numbers = _compute_element_numbers(unique_elements)

    end()

    return form_datas, unique_elements, element_numbers, unique_coordinate_elements


def _compute_element_numbers(elements):
    "Build map from elements to element numbers."
    element_numbers = {}
    for (i, element) in enumerate(elements):
        element_numbers[element] = i
    return element_numbers


def _analyze_form(form, parameters):
    "Analyze form, returning form data."

    # Check that form is not empty
    if form.empty():
        error("Form (%s) seems to be zero: cannot compile it." % str(form))

    # Hack to override representation with environment variable
    forced_r = os.environ.get("FFC_FORCE_REPRESENTATION")
    if forced_r:
        warning("representation:    forced by $FFC_FORCE_REPRESENTATION to '%s'" % forced_r)
        r = forced_r
        r = "legacy" if r in ["quadrature", "tensor"] else r
    else:
        # Check representation parameters to figure out how to preprocess
        r = _extract_representation_family(form, parameters)
    debug("Preprocessing form using '%s' representation family." % r)

    # Compute form metadata
    if r == "uflacs":
        # Temporary workaround to let uflacs have a different preprocessing pipeline
        # than the legacy representations quadrature and tensor. This approach imposes
        # a limitation that e.g. uflacs and tensor representation cannot be mixed in the same form.
        from ufl.classes import Jacobian
        form_data = compute_form_data(form,
                                      do_apply_function_pullbacks=True,
                                      do_apply_integral_scaling=True,
                                      do_apply_geometry_lowering=True,
                                      preserve_geometry_types=(Jacobian,),
                                      do_apply_restrictions=True,
                                      )
    elif r == "tsfc":
        try:
            # TSFC provides compute_form_data wrapper using correct kwargs
            from tsfc.ufl_utils import compute_form_data as tsfc_compute_form_data
        except ImportError:
            error("Failed to import tsfc.ufl_utils.compute_form_data when asked "
                  "for tsfc representation.")
        form_data = tsfc_compute_form_data(form)
    elif r == "legacy":
        # quadrature or tensor representation
        form_data = compute_form_data(form)
    else:
        error("Unexpected representation family '%s' for form preprocessing." % r)

    info("")
    info(str(form_data))

    # Attach integral meta data
    _attach_integral_metadata(form_data, r, parameters)
    _validate_representation_choice(form_data, r)

    return form_data


def _extract_representation_family(form, parameters):
    """Return 'uflacs', 'tsfc' or 'legacy', or raise error. This
    takes care of (a) compatibility between representations due to
    differences in preprocessing, (b) choosing uflacs for
    higher-order geometries.

    NOTE: Final representation is picked later by
    ``_determine_representation``.
    """
    # Fetch all representation choice requests from metadata
    representations = set(integral.metadata().get("representation", "auto")
                          for integral in form.integrals())

    # If auto is present, add parameters value (which may still be auto)
    # and then remove auto so there's no auto in the set
    if "auto" in representations:
        representations.add(parameters["representation"])
        representations.remove("auto")

    # Translate quadrature/tensor to legacy
    for r in list(representations):
        if r in ['quadrature', 'tensor']:
            representations.remove(r)
            representations.add('legacy')

    # Sanity check
    ffc_assert(len(representations.intersection((
        'quadrature', 'tensor', 'auto', None))) == 0,
        "Unexpected representation family candidates '%s'." % representations)

    # No representations requested, find compatible representations
    compatible = _find_compatible_representations(form.integrals(), [])
    # Translate quadrature/tensor to legacy
    for r in list(compatible):
        if r in ['quadrature', 'tensor']:
            compatible.remove(r)
            compatible.add('legacy')

    if len(representations) == 1:
        r = representations.pop()
        if r not in compatible:
            error("Representation family %s is not compatible with this form (try one of %s)" % (r, sorted(compatible)))
        return r
    elif len(representations) == 0:
        if len(compatible) == 1:
            # If only one compatible, use it
            return compatible.pop()
        else:
            # Default to uflacs
            # NOTE: Need to pick the same default as in _auto_select_representation
            return "uflacs"
    else:
        # Don't tolerate user requests for mixing old and new representation
        # families in same form due to restrictions in preprocessing
        assert len(representations) > 1
        error("Cannot mix legacy (quadrature, tensor), uflacs, or tsfc "
              "representation in single form.")


def _validate_representation_choice(form_data,
                                    preprocessing_representation_family):
    """Check that effective representations

    * do not mix legacy, uflacs and tsfc,
    * implement higher-order geometry,
    * match employed preprocessing strategy.

    This function is final check that everything is compatible due
    to the mess in this file. Better safe than sorry...
    """
    # Fetch all representations from integral metadata
    # (this should by now be populated with parameters or defaults instead of auto)
    representations = set()
    for ida in form_data.integral_data:
        representations.add(ida.metadata["representation"])
        for integral in ida.integrals:
            representations.add(integral.metadata()["representation"])

    # Translate quadrature/tensor to legacy
    for r in list(representations):
        if r in ['quadrature', 'tensor']:
            representations.remove(r)
            representations.add('legacy')

    # No integrals
    if len(representations) == 0:
        return

    # Require unique family; allow legacy only with affine meshes
    if len(representations) != 1:
        error("Failed to extract unique representation family. "
              "Got '%s'." % representations)

    if _has_higher_order_geometry(form_data.preprocessed_form):
        ffc_assert('legacy' not in representations,
            "Did not expect legacy representation for higher-order geometry.")

    # Check preprocessing strategy
    ffc_assert(preprocessing_representation_family in representations,
        "Form has been preprocessed using '%s' representaion family, "
        "while '%s' representations have been set for integrals."
        % (preprocessing_representation_family, representations))


def _has_custom_integrals(o):
    if isinstance(o, Integral):
        return o.integral_type() in custom_integral_types
    elif isinstance(o, Form):
        return any(_has_custom_integrals(itg) for itg in o.integrals())
    elif isinstance(o, (list, tuple)):
        return any(_has_custom_integrals(itg) for itg in o)
    else:
        raise NotImplementedError


def _has_higher_order_geometry(o):
    if isinstance(o, Integral):
        P1 = VectorElement("P", o.ufl_domain().ufl_cell(), 1)
        return o.ufl_domain().ufl_coordinate_element() != P1
    elif isinstance(o, Form):
        P1 = VectorElement("P", o.ufl_cell(), 1)
        return any(d.ufl_coordinate_element() != P1 for d in o.ufl_domains())
    elif isinstance(o, (list, tuple)):
        return any(_has_higher_order_geometry(itg) for itg in o)
    else:
        raise NotImplementedError


def _extract_common_quadrature_degree(integral_metadatas):
    # Check that quadrature degree is the same
    quadrature_degrees = [md["quadrature_degree"] for md in integral_metadatas]
    for d in quadrature_degrees:
        if not isinstance(d, int):
            error("Invalid non-integer quadrature degree %s" % (str(d),))
    qd = max(quadrature_degrees)
    if not all_equal(quadrature_degrees):
        # FIXME: Shouldn't we raise here?
        # TODO: This may be loosened up without too much effort,
        # if the form compiler handles mixed integration degree,
        # something that most of the pipeline seems to be ready for.
        info("Quadrature degree must be equal within each sub domain, using degree %d." % qd)
    return qd


def _autoselect_quadrature_degree(integral_metadata, integral, form_data):
    # Automatic selection of quadrature degree
    qd = integral_metadata["quadrature_degree"]
    pd = integral_metadata["estimated_polynomial_degree"]

    # Special case: handling -1 as "auto" for quadrature_degree
    if qd in [-1, None]:
        qd = "auto"

    # TODO: Add other options here
    if qd == "auto":
        qd = pd
        info("quadrature_degree: auto --> %d" % qd)
    if isinstance(qd, int):
        if qd >= 0:
            info("quadrature_degree: %d" % qd)
        else:
            error("Illegal negative quadrature degree %s " % (qd,))
    else:
        error("Invalid quadrature_degree %s." % (qd,))

    tdim = integral.ufl_domain().topological_dimension()
    _check_quadrature_degree(qd, tdim)
    return qd


def _check_quadrature_degree(degree, top_dim):
    """Check that quadrature degree does not result in a unreasonable high
    number of integration points."""
    num_points = ((degree + 1 + 1) // 2)**top_dim
    if num_points >= 100:
        warning_blue("WARNING: The number of integration points for each cell will be: %d" % num_points)
        warning_blue("         Consider using the option 'quadrature_degree' to reduce the number of points")


def _extract_common_quadrature_rule(integral_metadatas):
    # Check that quadrature rule is the same
    # (To support mixed rules would be some work since num_points is
    #  used to identify quadrature rules in large parts of the pipeline)
    quadrature_rules = [md["quadrature_rule"] for md in integral_metadatas]
    if all_equal(quadrature_rules):
        qr = quadrature_rules[0]
    else:
        qr = "canonical"
        # FIXME: Shouldn't we raise here?
        info("Quadrature rule must be equal within each sub domain, using %s rule." % qr)
    return qr


def _autoselect_quadrature_rule(integral_metadata, integral, form_data):
    # Automatic selection of quadrature rule
    qr = integral_metadata["quadrature_rule"]
    if qr in ["auto", None]:
        # Just use default for now.
        qr = "default"
        info("quadrature_rule:   auto --> %s" % qr)
    elif qr in ("default", "canonical", "vertex"):
        info("quadrature_rule:   %s" % qr)
    else:
        info("Valid choices are 'default', 'canonical', 'vertex', and 'auto'.")
        error("Illegal choice of quadrature rule for integral: " + str(qr))
    # Return automatically determined quadrature rule
    return qr


def _determine_representation(integral_metadatas, ida, form_data, form_r_family, parameters):
    "Determine one unique representation considering all integrals together."

    # Extract unique representation among these single-domain integrals
    # (Generating code with different representations within a
    # single tabulate_tensor is considered not worth the effort)
    representations  = set(md["representation"] for md in integral_metadatas
                           if md["representation"] != "auto")
    optimize_values  = set(md["optimize"] for md in integral_metadatas)
    precision_values = set(md["precision"] for md in integral_metadatas)

    if len(representations) > 1:
        error("Integral representation must be equal within each sub domain or 'auto', got %s." % (str(sorted(str(v) for v in representations)),))
    if len(optimize_values) > 1:
        error("Integral 'optimize' metadata must be equal within each sub domain or not set, got %s." % (str(sorted(str(v) for v in optimize_values)),))
    if len(precision_values) > 1:
        error("Integral 'precision' metadata must be equal within each sub domain or not set, got %s." % (str(sorted(str(v) for v in precision_values)),))

    # The one and only non-auto representation found, or get from parameters
    r, = representations  or (parameters["representation"],)
    o, = optimize_values  or (parameters["optimize"],)
    # FIXME: Default param value is zero which is not interpreted well by tsfc!
    p, = precision_values or (parameters["precision"],)

    # If it's still auto, try to determine which representation is best for these integrals
    if r == "auto":
        # Find representations compatible with these integrals
        compatible = _find_compatible_representations(ida.integrals,
                                                      form_data.unique_sub_elements)
        # Pick the one compatible or default to uflacs
        if len(compatible) == 0:
            error("Found no representation capable of compiling this form.")
        elif len(compatible) == 1:
            r, = compatible
        else:
            # NOTE: Need to pick the same default as in _extract_representation_family
            if form_r_family == "uflacs":
                r = "uflacs"
            elif form_r_family == "tsfc":
                r = "tsfc"
            elif form_r_family == "legacy":
                if "tensor" not in compatible:
                    r = "quadrature"
                elif "quadrature" not in compatible:
                    r = "tensor"
                else:
                    r = "tensor"
                    # Use quadrature if tensor representation is not possible or too costly
                    for integral in ida.integrals:
                        tensor_cost = estimate_cost(integral, form_data.function_replace_map)
                        debug("Estimated cost of tensor representation: " + str(tensor_cost))
                        if tensor_cost == -1 or tensor_cost > 3:
                            r = "quadrature"
                            break
            else:
                error("Invalid form representation family %s." % (form_r_family,))
        info("representation:    auto --> %s" % r)
    else:
        info("representation:    %s" % r)

    if p is None:
        p = default_precision

    # Hack to override representation with environment variable
    forced_r = os.environ.get("FFC_FORCE_REPRESENTATION")
    if forced_r:
        r = forced_r
        warning("representation:    forced by $FFC_FORCE_REPRESENTATION to '%s'" % r)
        return r, o, p

    return r, o, p


def _attach_integral_metadata(form_data, form_r_family, parameters):
    "Attach integral metadata"
    # TODO: A nicer data flow would avoid modifying the form_data at all.

    # Parameter values which make sense "per integrals" or "per integral"
    metadata_keys = (
        "representation",
        "optimize",
        # TODO: Could have finer optimize (sub)parameters here later
        "precision",
        # NOTE: We don't pass precision to quadrature and tensor, it's not
        #       worth resolving set_float_formatting hack for (almost)
        #       deprecated backends
        "quadrature_degree",
        "quadrature_rule",
    )

    # Get defaults from parameters
    metadata_parameters = {key: parameters[key] for key in metadata_keys if key in parameters}

    # Iterate over integral collections
    quad_schemes = []
    for ida in form_data.integral_data:
        # Iterate over integrals

        # Start with default values of integral metadata
        # (these will be either the FFC defaults, globally modified defaults,
        #  or overrides explicitly passed by the user to e.g. assemble())
        integral_metadatas = [copy.deepcopy(metadata_parameters)
                              for integral in ida.integrals]

        # Update with integral specific overrides
        for i, integral in enumerate(ida.integrals):
            integral_metadatas[i].update(integral.metadata() or {})

        # Determine representation, must be equal for all integrals on same subdomain
        r, o, p = _determine_representation(integral_metadatas, ida, form_data, form_r_family, parameters)
        for i, integral in enumerate(ida.integrals):
            integral_metadatas[i]["representation"] = r
            integral_metadatas[i]["optimize"] = o
            integral_metadatas[i]["precision"] = p
        ida.metadata["representation"] = r
        ida.metadata["optimize"] = o
        ida.metadata["precision"] = p

        # Determine automated updates to metadata values
        for i, integral in enumerate(ida.integrals):
            qr = _autoselect_quadrature_rule(integral_metadatas[i], integral, form_data)
            qd = _autoselect_quadrature_degree(integral_metadatas[i], integral, form_data)
            integral_metadatas[i]["quadrature_rule"] = qr
            integral_metadatas[i]["quadrature_degree"] = qd

        # Extract common metadata for integral collection
        qr = _extract_common_quadrature_rule(integral_metadatas)
        qd = _extract_common_quadrature_degree(integral_metadatas)
        ida.metadata["quadrature_rule"] = qr
        ida.metadata["quadrature_degree"] = qd

        # Add common num_cells (I believe this must be equal but I'm not that into this work)
        num_cells = set(md.get("num_cells") for md in integral_metadatas)
        if len(num_cells) != 1:
            error("Found integrals with different num_cells metadata on same subdomain: %s" % (str(list(num_cells)),))
        num_cells, = num_cells
        ida.metadata["num_cells"] = num_cells

        # Reconstruct integrals to avoid modifying the input integral,
        # which would affect the signature computation if the
        # integral was used again in the user program.
        # Modifying attributes of form_data.integral_data is less problematic
        # since it's lifetime is internal to the form compiler pipeline.
        for i, integral in enumerate(ida.integrals):
            ida.integrals[i] = integral.reconstruct(metadata=integral_metadatas[i])

        # Collect all quad schemes
        quad_schemes.extend([md["quadrature_rule"] for md in integral_metadatas])

    # Validate consistency of schemes for QuadratureElements
    # TODO: Can loosen up this a bit, only needs to be consistent
    # with the integrals that the elements are used in
    _validate_quadrature_schemes_of_elements(quad_schemes, form_data.unique_sub_elements)


def _validate_quadrature_schemes_of_elements(quad_schemes, elements):
    # Update scheme for QuadratureElements
    if quad_schemes and all_equal(quad_schemes):
        scheme = quad_schemes[0]
    else:
        scheme = "canonical"
        info("Quadrature rule must be equal within each sub domain, using %s rule." % scheme)
    for element in elements:
        if element.family() == "Quadrature":
            qs = element.quadrature_scheme()
            if qs != scheme:
                error("Quadrature element must have specified quadrature scheme (%s) equal to the integral (%s)." % (qs, scheme))


def _get_sub_elements(element):
    "Get sub elements."
    sub_elements = [element]
    if isinstance(element, MixedElement):
        for e in element.sub_elements():
            sub_elements += _get_sub_elements(e)
    elif isinstance(element, EnrichedElement):
        for e in element._elements:
            sub_elements += _get_sub_elements(e)
    return sub_elements


def _find_compatible_representations(integrals, elements):
    """
    Automatically select a suitable representation for integral.
    Note that the selection is made for each integral, not for
    each term. This means that terms which are grouped by UFL
    into the same integral (if their measures are equal) will
    necessarily get the same representation.
    """
    # All representations
    compatible = set(("uflacs", "quadrature", "tensor", "tsfc"))

    # Check for non-affine meshes
    if _has_higher_order_geometry(integrals):
        compatible &= set(("uflacs", "tsfc"))

    # Custom integrals
    if _has_custom_integrals(integrals):
        compatible &= set(("quadrature",))

    # Use quadrature for vertex integrals
    if any(integral.integral_type() == "vertex" for integral in integrals):
        # TODO: Test with uflacs, I think this works fine now:
        compatible &= set(("quadrature", "uflacs", "tsfc"))

    # Get ALL sub elements, needed to check for restrictions of EnrichedElements.
    sub_elements = []
    for e in elements:
        sub_elements += _get_sub_elements(e)

    # Use quadrature representation if we have a quadrature element
    if any(e.family() == "Quadrature" for e in sub_elements):
        # TODO: Test with uflacs, might need a little adjustment:
        compatible &= set(("quadrature", "uflacs", "tsfc"))

    return compatible



# -*- coding: utf-8 -*-
# Copyright (C) 2010 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

import pytest
import time

# FFC modules
from ffc.quadrature.symbolics import *
from ffc.quadrature.sumobj import _group_fractions
from ffc.cpp import format, set_float_formatting
from ffc.parameters import FFC_PARAMETERS
set_float_formatting(FFC_PARAMETERS['precision'])


def testNotFinished():
    "Stuff that would be nice to implement."

    f_1 = format["float"](1)
    f_2 = format["float"](2)
    f_4 = format["float"](4)
    f_8 = format["float"](8)

    f0 = FloatValue(4)
    f1 = FloatValue(2)
    f2 = FloatValue(8)
    s0 = Symbol("x", GEO)
    s1 = Symbol("y", GEO)
    s2 = Symbol("z", GEO)
    a = Symbol("a", GEO)
    b = Symbol("b", GEO)
    c = Symbol("c", GEO)

    # Aux. expressions
    p0 = Product([f1, s0])
    p1 = Product([f2, s1])
    p2 = Product([s0, s1])

    F0 = Fraction(f0, s0)

    S0 = Sum([p0, p1])
    S1 = Sum([s0, p2])
    S2 = Sum([FloatValue(1), s1])
    S3 = Sum([F0, F0])

    # Thing to be implemented
    e0 = f0 / S0
    e1 = s0 / S1
    e2 = S2 / S1
    e3 = _group_fractions(S3)
    e4 = Sum([Fraction(f1 * s0, a * b * c), Fraction(s0, a * b)]).expand().reduce_ops()

    # Tests that pass the current implementation
    assert str(e0) == '%s/(%s*x + %s*y)' % (f_4, f_2, f_8)
    assert str(e1) == 'x/(x + x*y)'
    assert str(e2) == '(%s + y)/(x + x*y)' % f_1
    assert str(e3) == '%s/x' % f_8
    assert str(e4) == 'x*(%s/(a*b) + %s/(a*b*c))' % (f_1, f_2)

    # Tests that should pass in future implementations (change NotEqual to Equal)
    assert str(e0) != '%s/(x + %s*y)' % (f_2, f_4)
    assert str(e1) != '%s/(%s + y)' % (f_1, f_1)
    assert str(e2) != '%s/x' % f_1
    assert str(e4) != 'x*(%s/c + %s)/(a*b)' % (f_2, f_1)

    # TODO: Would it be a good idea to reduce expressions
    # wrt. var_type without first expanding?
    E0 = Product([Sum([Product([Symbol('B0', BASIS), Product([Symbol('B1', BASIS), Sum([s0]), Sum([s0])])]),
                       Product([Symbol('B0', BASIS), Symbol('B1', BASIS)])])])
    Er0 = E0.reduce_vartype(BASIS)
    Ex0 = E0.expand().reduce_vartype(BASIS)
    assert Ex0[0][1] != Er0[0][1].expand()

    # Both of these reductions should work at the same time
    # 1) 2/(x/(a+b) + y/(a+b)) --> 2(a+b)/(x+y)
    # 2) 2/(x + y/(a+b)) --> no reduction, or if divisions are more expensive
    # 3) 2/(x + y/(a+b)) --> 2(a+b)/((a+b)x + y)



# -*- coding: utf-8 -*-
"This script plots the results found in bench.log."

# Copyright (C) 2010 Anders Logg
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# First added:  2010-05-13
# Last changed: 2010-05-13

from pylab import *

# Read logfile
results = {}
try:
    output = open("bench.log").read()
except:
    output = open("results/bench.log").read()
for line in output.split("\n"):
    if "," not in line: continue
    test_case, test_option, timing = [w.strip() for w in line.split(",")]
    try:
        form, degree = test_case.split("_")
    except:
        form, dim, degree = test_case.split("_")
        form = form + "_" + dim
    if form not in results:
        results[form] = {}
    if test_option not in results[form]:
        results[form][test_option] = ([], [])
    results[form][test_option][0].append(int(degree))
    results[form][test_option][1].append(float(timing))

# Plot results
forms = sorted([form for form in results])
test_options = ["-r quadrature", "-r quadrature -O", "-r tensor", "-r tensor -O"]
bullets = ["x-", "o-", "*-", "s-"]
for (i, form) in enumerate(forms):
    figure(i)

    # Plot timings
    subplot(121)
    for (j, test_option) in enumerate(test_options):
        q, t = results[form][test_option]
        semilogy(q, t, bullets[j])
        hold(True)
    a = list(axis()); a[-1] = 50.0*a[-1]; axis(a);
    legend(test_options, loc="upper left")
    grid(True)
    xlabel('degree')
    ylabel(form)
    title('CPU time')

    # Plot speedups
    subplot(122)
    q0, t0 = results[form]["-r quadrature"]
    for (j, test_option) in enumerate(test_options):
        q, t = results[form][test_option]
        t = [t0[k] / t[k] for k in range(len(t))]
        semilogy(q, t, bullets[j])
        hold(True)
    a = list(axis()); a[-1] = 50.0*a[-1]; axis(a);
    legend(test_options, loc="upper left")
    grid(True)
    xlabel('degree')
    title("Speedup vs '-r quadrature'")

show()



# -*- coding: utf-8 -*-
# Copyright (C) 2004-2009 Anders Logg
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Marie E. Rognes, 2007
# Modified by Kristian B. Oelgaard, 2009
#
# First added:  2004-11-03
# Last changed: 2009-12-21

# FFC modules.
from ffc.log import debug

# FFC tensor representation modules.
from .monomialtransformation import MonomialIndex
from .multiindex import create_multiindex

class GeometryTensor:
    """
    This class represents the geometry tensor for a monomial term of a
    multilinear form.
    """

    def __init__(self, monomial):
        "Create geometry tensor for given monomial."

        # Save monomial data
        self.determinants = monomial.determinants
        self.coefficients = monomial.coefficients
        self.transforms = monomial.transforms

        # Extract indices
        secondary_indices = monomial.extract_unique_indices(MonomialIndex.SECONDARY)
        external_indices  = monomial.extract_unique_indices(MonomialIndex.EXTERNAL)

        # Create multiindices
        self.secondary_multi_index = create_multiindex(secondary_indices)
        self.external_multi_index  = create_multiindex(external_indices)

        debug("Secondary multi index: " + str(self.secondary_multi_index))
        debug("External multi index:  " + str(self.external_multi_index))



# -*- coding: utf-8 -*-
"""
Tests of algorithm for factorization of integrand w.r.t. Argument terms.
"""

from six.moves import xrange as range
from ufl import *
from ffc.uflacs.analysis.factorization import compute_argument_factorization

# TODO: Restructure these tests using py.test fixtures and parameterization?


def compare_compute_argument_factorization(SV, SV_deps, expected_AV, expected_FV, expected_IM):
    SV_targets = [len(SV) - 1]
    rank = max(len(k) for k in expected_IM.keys())
    argument_factorizations, modified_arguments, FV, FV_deps, FV_targets = \
        compute_argument_factorization(SV, SV_deps, SV_targets, rank)
    argument_factorization, = argument_factorizations

    AV = modified_arguments
    IM = argument_factorization

    assert AV == expected_AV
    if 0:
        for n in range(1, min(len(FV), len(expected_FV))):
            assert FV[:n] == expected_FV[:n]
    assert FV == expected_FV
    assert IM == expected_IM


def test_compute_argument_factorization():
    V = FiniteElement("CG", triangle, 1)
    u = TrialFunction(V)
    v = TestFunction(V)
    a, b, c, d, e, f, g = [Coefficient(V, count=k) for k in range(7)]

    zero = as_ufl(0.0)
    one = as_ufl(1.0)
    two = as_ufl(2)

    # Test workaround for hack in factorization:
    FVpre = [zero, one, two]
    offset = len(FVpre)

    # Test basic non-argument terminal
    SV = [f]
    SV_deps = [()]
    AV = []
    FV = FVpre + [f]
    IM = {(): 0 + offset}
    compare_compute_argument_factorization(SV, SV_deps, AV, FV, IM)

    # Test basic non-argument sum
    SV = [f, g, f + g]
    SV_deps = [(), (), (0, 1)]
    AV = []
    FV = FVpre + [f, g, f + g]
    IM = {(): 2 + offset}
    compare_compute_argument_factorization(SV, SV_deps, AV, FV, IM)

    # Test basic non-argument product
    SV = [f, g, f * g]
    SV_deps = [(), (), (0, 1)]
    AV = []
    FV = FVpre + [f, g, f * g]
    IM = {(): 2 + offset}
    compare_compute_argument_factorization(SV, SV_deps, AV, FV, IM)

    # Test basic single-argument-only expression
    SV = [v]
    SV_deps = [()]
    AV = [v]
    FV = FVpre + []
    IM = {(0,): 1}  # v == AV[0] * FV[1]
    compare_compute_argument_factorization(SV, SV_deps, AV, FV, IM)

    # Test basic coefficient-argument product
    SV = [f, v, f * v]
    SV_deps = [(), (), (0, 1)]
    AV = [v]
    FV = FVpre + [f]
    IM = {(0,): offset}  # f*v == AV[0] * FV[1]
    compare_compute_argument_factorization(SV, SV_deps, AV, FV, IM)

    # Test basic argument product
    SV = [u, v, u * v]
    SV_deps = [(), (), (0, 1)]
    AV = [v, u]  # Test function < trial function
    FV = FVpre + []
    IM = {(0, 1): 1}  # v*u == (AV[0] * AV[1]) * FV[1]
    compare_compute_argument_factorization(SV, SV_deps, AV, FV, IM)

    # Test coefficient-argument products
    SV = [u, f, v, (f * v), u * (f * v)]
    SV_deps = [(), (), (), (1, 2), (0, 3)]
    AV = [v, u]
    FV = FVpre + [f]
    IM = {(0, 1): 0 + offset}  # f*(u*v) == (AV[0] * AV[1]) * FV[2]
    compare_compute_argument_factorization(SV, SV_deps, AV, FV, IM)

    # Test more complex situation
    SV = [u, u.dx(0), v,  # 0..2
          a, b, c, d, e,  # 3..7
          a * u, b * u.dx(0),  # 8..9
          c * v, d * v,  # 10..11
          a * u + b * u.dx(0),  # 12
          c * v + d * v,  # 13
          e * (a * u + b * u.dx(0)),  # 14
          (e * (a * u + b * u.dx(0))) * (c * v + d * v),  # 15
          ]
    SV_deps = [(), (), (),
                    (), (), (), (), (),
                    (0, 3), (1, 4),
                    (2, 5), (2, 6),
                    (8, 9),
                    (10, 11),
                    (7, 12),
                    (13, 14),
                    ]
    AV = [v, u, u.dx(0)]
    FV = FVpre + [a, b, c, d, e,  # 0..5
                  c + d,  # 6, introduced by SV[13]
                  e * a,  # 7, introduced by SV[14]
                  e * b,  # 8, introduced by SV[14]
                  (e * a) * (c + d),  # 9
                  (e * b) * (c + d),  # 10
                  ]
    IM = {(0, 1): 8 + offset,  # (a*e)*(c+d)*(u*v) == (AV[0] * AV[2]) * FV[13]
          (0, 2): 9 + offset}  # (b*e)*(c+d)*(u.dx(0)*v) == (AV[1] * AV[2]) * FV[12]
    compare_compute_argument_factorization(SV, SV_deps, AV, FV, IM)



# -*- coding: utf-8 -*-
"Code generation for interpolate_vertex_values."

# Copyright (C) 2009 Marie E. Rognes
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Kristian B. Oelgaard 2010
# Modified by Lizao Li 2015, 2016
#
# Last changed: 2016-08-17

from six import string_types
from ffc.cpp import format, remove_unused

# Extract code manipulation formats
inner = format["inner product"]
component = format["component"]
assign = format["assign"]
multiply = format["multiply"]

# Extract formats for the Jacobians
J = format["J"]
Jinv = format["inv(J)"]
invdetJ = format["inverse"](format["det(J)"](None))

f_dof_values = format["argument dof values"]
f_vertex_values = format["argument vertex values"]


def interpolate_vertex_values(ir):
    "Generate code for interpolate_vertex_values."

    # Handle unsupported elements.
    if isinstance(ir, string_types):
        return format["exception"]("interpolate_vertex_values: %s" % ir)

    # Add code for Jacobian if necessary
    code = []
    gdim = ir["geometric_dimension"]
    tdim = ir["topological_dimension"]
    if ir["needs_jacobian"]:

        # Generate code for basic geometric quantities
        code.append(format["compute_jacobian"](tdim, gdim))
        code.append("")
        code.append(format["compute_jacobian_inverse"](tdim, gdim))
        if ir["needs_oriented"]:
            code.append("")
            code.append(format["orientation"](tdim, gdim))

    # Compute total value dimension for (mixed) element
    total_dim = ir["physical_value_size"]

    # Generate code for each element
    value_offset = 0
    space_offset = 0
    for data in ir["element_data"]:
        # Add vertex interpolation for this element
        code.append(format["comment"]("Evaluate function and change variables"))
        code.append(_interpolate_vertex_values_element(data, gdim, tdim,
                                                       total_dim,
                                                       value_offset,
                                                       space_offset))

        # Update offsets for value- and space dimension
        value_offset += data["physical_value_size"]
        space_offset += data["space_dim"]

    # Remove unused variables. (Not tracking set of used variables in
    # order to keep this code clean. Since generated code is of
    # limited size, this should be ok.)
    clean_code = remove_unused("\n".join(code))
    return clean_code


def _interpolate_vertex_values_element(data, gdim, tdim, total_value_size,
                                       value_offset=0, space_offset=0):

    # Extract vertex values for all basis functions
    vertex_values = data["basis_values"]
    value_size = data["physical_value_size"]
    space_dim = data["space_dim"]
    mapping = data["mapping"]

    # Map basis values according to element mapping. Assumes single
    # mapping for each (non-mixed) element
    change_of_variables = _change_variables(data["mapping"], gdim, tdim,
                                            space_dim)

    # Create code for each value dimension:
    code = []
    for k in range(value_size):

        # Create code for each vertex x_j
        for (j, values_at_vertex) in enumerate(vertex_values):

            if value_size == 1:
                values_at_vertex = [values_at_vertex]

            # Map basis functions using appropriate mapping
            components = change_of_variables(values_at_vertex, k)

            # Contract coefficients and basis functions
            dof_values = [component(f_dof_values, i + space_offset)
                          for i in range(space_dim)]

            value = inner(dof_values, components)

            # Assign value to correct vertex
            index = j * total_value_size + (k + value_offset)
            code.append(assign(component(f_vertex_values, index), value))

    return "\n".join(code)


def _change_variables(mapping, gdim, tdim, space_dim):
    """

    How to map a field G from the reference domain to a physical
    domain: For the converse approach -- see evaluatedof.py

    Let g be a field defined on the reference domain T_0 (of
    dimension tdim) with reference coordinates X. Let T be a a
    physical domain (of dimension gdim) with coordinates x. Assume
    that F: T_0 -> T such that

      x = F(X)

    Let J be the Jacobian of F, i.e J = dx/dX and let K denote the
    (pseudo)-inverse of the Jacobian K = J^{-1}. Note that J is gdim x
    tdim, and conversely K is tdim x gdim.  Then we (currently) have
    the following four types of mappings:

    'affine' mapping for G:

      g(x) = G(X)

    For vector fields G:

    'contravariant piola' mapping for f:

      g(x) = 1.0/det(J) J G(X)   i.e   g_i(x) = 1.0/det(J) J_ij G_j(X)

    'covariant piola' mapping for f:

      g(x) = K^T G(X)              i.e   g_i(x) = K^T_ij G_j(X) = K_ji G_j(X)

    'double covariant piola' mapping for f:

      g_il(x) = K_{ji} G_{jk} K_{kl}

    'double contravariant piola' mapping for g:

      g_il(x) = (det(J))^(-2) J_ij G_jk(X) J_lk

    """

    if mapping is "affine":
        change_of_variables = lambda G, i: G[i]
    elif mapping == "contravariant piola":
        change_of_variables = lambda G, i: [multiply([invdetJ, inner([J(i, j, gdim, tdim) for j in range(tdim)],
                                                                     [G[j][index] for j in range(tdim)])])
                                            for index in range(space_dim)]
    elif mapping == "covariant piola":
        change_of_variables = lambda G, i: [inner([Jinv(j, i, tdim, gdim) for j in range(tdim)],
                                                  [G[j][index] for j in range(tdim)])
                                            for index in range(space_dim)]
    elif mapping == "double covariant piola":
        change_of_variables = lambda G, i: [
            inner([inner([Jinv(j, i // tdim, tdim, gdim) for j in range(tdim)],
                         [G[j][k][index] for j in range(tdim)])
                   for k in range(tdim)],
                  [Jinv(k, i % tdim, tdim, gdim) for k in range(tdim)])
            for index in range(space_dim)]
    elif mapping == "double contravariant piola":
        change_of_variables = lambda G, i: [
            multiply([invdetJ, invdetJ, inner(
                [inner([J(i // tdim, j, tdim, gdim) for j in range(tdim)],
                         [G[j][k][index] for j in range(tdim)])
                   for k in range(tdim)],
                  [J(i % tdim, k, tdim, gdim) for k in range(tdim)])])
            for index in range(space_dim)]
    else:
        raise Exception("No such mapping: %s accepted" % mapping)
    return change_of_variables



# -*- coding: utf-8 -*-
# Copyright (C) 2010 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

import pytest
import time

# FFC modules
from ffc.quadrature.symbolics import *
from ffc.cpp import format, set_float_formatting
from ffc.parameters import FFC_PARAMETERS
set_float_formatting(FFC_PARAMETERS['precision'])


def testReduceVarType():
    f1 = FloatValue(1)
    f2 = FloatValue(2)
    f3 = FloatValue(3)
    f5 = FloatValue(5)
    fm4 = FloatValue(-4)

    B0 = Symbol("B0", BASIS)
    B1 = Symbol("B1", BASIS)
    Bm4 = Product([fm4, B1])
    B5 = Product([f5, B0])

    I0 = Symbol("I0", IP)
    I1 = Symbol("I1", IP)
    I2 = Symbol("I2", IP)
    I5 = Product([f5, I0])

    G0 = Symbol("G0", GEO)
    G1 = Symbol("G1", GEO)
    G2 = Symbol("G2", GEO)
    G3 = Product([f3, G0])

    C0 = Symbol("C0", CONST)
    C2 = Product([f2, C0])

    p0 = Product([B0, I5])
    p1 = Product([B0, B1])

    S0 = Sum([B0, I5])
    S1 = Sum([p0, p1])
    S2 = Sum([B0, B1])
    S3 = Sum([B0, p0])
    S4 = Sum([f5, p0])
    S5 = Sum([I0, G0])

    F0 = Fraction(B0, I5).expand()
    F1 = Fraction(p1, I5).expand()
    F2 = Fraction(G3, S2).expand()
    F3 = Fraction(G3, S3).expand()
    F4 = Fraction(I1, Sum([I1, I0]))
    F5 = Fraction(S5, I1)
    F6 = Fraction(I0,
                  Sum([
                      Fraction(Sum([I0, I1]), Sum([G0, G1])),
                      Fraction(Sum([I1, I2]), Sum([G1, G2])),
                  ]))

    r0 = B0.reduce_vartype(BASIS)
    r1 = B0.reduce_vartype(CONST)

    rp0 = p0.reduce_vartype(BASIS)
    rp1 = p0.reduce_vartype(IP)
    rp2 = p1.reduce_vartype(BASIS)
    rp3 = p1.reduce_vartype(GEO)

    rs0 = S0.reduce_vartype(BASIS)
    rs1 = S0.reduce_vartype(IP)
    rs2 = S1.reduce_vartype(BASIS)
    rs3 = S4.reduce_vartype(BASIS)
    rs4 = S4.reduce_vartype(CONST)

    rf0 = F0.reduce_vartype(BASIS)
    rf1 = F1.reduce_vartype(BASIS)
    rf2 = F0.reduce_vartype(IP)
    rf3 = F2.reduce_vartype(BASIS)
    rf4 = F3.reduce_vartype(BASIS)
    rf5 = F4.reduce_vartype(IP)
    rf6 = F5.reduce_vartype(IP)
    rf7 = F6.reduce_vartype(IP)

    assert [(B0, f1)] == r0
    assert [((), B0)] == r1

    assert [(B0, I5)] == rp0
    assert [(I0, B5)] == rp1
    assert [(p1, f1)] == rp2
    assert [((), p1)] == rp3

    assert ((), I5) == rs0[0]
    assert (B0, f1) == rs0[1]
    assert (I0, f5) == rs1[1]
    assert ((), B0) == rs1[0]
    assert (Product([B0, B1]), f1) == rs2[1]
    assert (B0, I5) == rs2[0]
    assert ((), f5) == rs3[0]
    assert (B0, I5) == rs3[1]
    assert (f5, Sum([f1, Product([B0, I0])])) == rs4[0]

    assert [(B0, Fraction(FloatValue(0.2), I0))] == rf0
    assert [(Product([B0, B1]), Fraction(FloatValue(0.2), I0))] == rf1
    assert [(Fraction(f1, I0), Product([FloatValue(0.2), B0]))] == rf2
    assert [(Fraction(f1, S2), G3)] == rf3
    assert [(Fraction(f1, B0), Fraction(G3, Sum([I5, f1])))] == rf4
    assert F4 == rf5[0][0]
    assert FloatValue(1) == rf5[0][1]
    assert Fraction(I0, I1) == rf6[1][0]
    assert f1 == rf6[1][1]
    assert Fraction(f1, I1) == rf6[0][0]
    assert G0 == rf6[0][1]
    assert F6 == rf7[0][0]
    assert f1 == rf7[0][1]

    expr = Sum([Symbol('W1', GEO), Fraction(Symbol('det', GEO), Sum([Symbol('F0', IP), Symbol('K_11', GEO)]))])
    red = expr.expand().reduce_vartype(IP)
    vals = []
    for ip in red:
        ip_dec, geo = ip
        if ip_dec and geo:
            vals.append(Product([ip_dec, geo]))
        elif geo:
            vals.append(geo)
        elif ip_dec:
            vals.append(ip_dec)
    comb = Sum(vals).expand()
    K_11 = 1.4
    F0 = 1.5
    W1 = 1.9
    det = 2.1
    assert round(eval(str(expr)) - eval(str(comb)), 10) == 0.0



# -*- coding: utf-8 -*-
# Copyright (C) 2011 Marie E. Rognes
#
# This file is part of DOLFIN.
#
# DOLFIN is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DOLFIN is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with DOLFIN. If not, see <http://www.gnu.org/licenses/>.
#
# Based on original implementation by Martin Alnes and Anders Logg
#
# Modified by Anders Logg 2015
#
# Last changed: 2016-03-02

from .includes import snippets
from .functionspace import *
from .goalfunctional import generate_update_ec

__all__ = ["generate_form"]


def generate_form(form, classname, error_control):
    """Generate dolfin wrapper code associated with a form including
    code for function spaces used in form and typedefs

    @param form:
        A UFCFormNames instance
    @param classname
        Name of Form class.
    """

    blocks = []

    # Generate code for Form_x_FunctionSpace_y subclasses
    wrap = apply_function_space_template
    blocks += [wrap("%s_FunctionSpace_%d" % (classname, i),
                    form.ufc_finite_element_classnames[i],
                    form.ufc_dofmap_classnames[i]) for i in range(form.rank)]

    # Generate code for Form_x_MultiMeshFunctionSpace_y subclasses
    wrap = apply_multimesh_function_space_template
    if not error_control:  # FIXME: Issue #91
        blocks += [wrap("%s_MultiMeshFunctionSpace_%d" % (classname, i),
                        "%s_FunctionSpace_%d" % (classname, i),
                        form.ufc_finite_element_classnames[i],
                        form.ufc_dofmap_classnames[i]) for i in range(form.rank)]

    # Add typedefs CoefficientSpace_z -> Form_x_FunctionSpace_y
    blocks += ["typedef CoefficientSpace_%s %s_FunctionSpace_%d;\n"
               % (form.coefficient_names[i], classname, form.rank + i)
               for i in range(form.num_coefficients)]

    # Generate Form subclass
    blocks += [generate_form_class(form, classname, error_control)]

    # Return code
    return "\n".join(blocks)


def generate_form_class(form, classname, error_control):
    "Generate dolfin wrapper code for a single Form class."

    # Generate constructors
    constructors = generate_form_constructors(form, classname)
    multimesh_constructors = generate_multimesh_form_constructors(form, classname)

    # Generate data for coefficient assignments
    (number, name) = generate_coefficient_map_data(form)

    # Generate typedefs for FunctionSpace subclasses for Coefficients
    typedefs = ["  // Typedefs", generate_typedefs(form, classname, error_control), ""]

    # Member variables for coefficients
    members = ["  dolfin::CoefficientAssigner %s;" % coefficient
               for coefficient in form.coefficient_names]
    if form.superclassname == "GoalFunctional":
        members += [generate_update_ec(form)]

    # Group typedefs and members together for inserting into template
    additionals = "\n".join(typedefs + ["  // Coefficients"] + members)

    # Wrap functions in class body
    code = ""
    code += apply_form_template(classname, constructors, number, name,
                                additionals, form.superclassname)
    code += "\n"
    if not error_control:
        code += apply_multimesh_form_template(classname, multimesh_constructors, number, name,
                                              additionals, form.superclassname)

    # Return code
    return code


def generate_coefficient_map_data(form):
    """Generate data for code for the functions
    Form::coefficient_number and Form::coefficient_name."""

    # Write error if no coefficients
    if form.num_coefficients == 0:
        message = '''\
dolfin::dolfin_error("generated code for class %s",
                         "access coefficient data",
                         "There are no coefficients");''' % form.superclassname
        num = "\n    %s\n    return 0;" % message
        name = '\n    %s\n    return "unnamed";' % message
        return (num, name)

    # Otherwise create switch
    ifstr = "if "
    num = ""
    name = '    switch (i)\n    {\n'
    for i, coeff in enumerate(form.coefficient_names):
        num += '    %s(name == "%s")\n      return %d;\n' % (ifstr, coeff, i)
        name += '    case %d:\n      return "%s";\n' % (i, coeff)
        ifstr = 'else if '

    # Create final return
    message = '''\
dolfin::dolfin_error("generated code for class %s",
                         "access coefficient data",
                         "Invalid coefficient");''' % form.superclassname
    num += "\n    %s\n    return 0;" % message
    name += '    }\n\n    %s\n    return "unnamed";' % message

    return (num, name)


def generate_form_constructors(form, classname):
    """Generate the dolfin::Form constructors for different
    combinations of references/shared pointers etc."""

    # FIXME: This can be simplied now that we don't create reference version

    coeffs = ("shared_ptr_coefficient",)
    spaces = ("shared_ptr_space",)

    # Treat functionals a little special
    if form.rank == 0:
        spaces = ("shared_ptr_mesh",)

    # Generate permutations of constructors
    constructors = []
    for space in spaces:
        constructors += [generate_constructor(form, classname, space)]
        if form.num_coefficients > 0:
            constructors += [generate_constructor(form, classname, space, coeff)
                             for coeff in coeffs]

    # Return joint constructor code
    return "\n\n".join(constructors)


def generate_multimesh_form_constructors(form, classname):
    """Generate the dolfin::MultiMeshForm constructors for different
    combinations of references/shared pointers etc."""

    # FIXME: This can be simplied now that we don't create reference version

    coeffs = ("shared_ptr_coefficient",)
    spaces = ("multimesh_shared_ptr_space",)

    # Treat functionals a little special
    if form.rank == 0:
        spaces = ("shared_ptr_multimesh",)

    # Generate permutations of constructors
    constructors = []
    for space in spaces:
        constructors += [generate_multimesh_constructor(form, classname, space)]
        if form.num_coefficients > 0:
            constructors += [generate_multimesh_constructor(form, classname,
                                                            space, coeff)
                             for coeff in coeffs]

    # Return joint constructor code
    return "\n\n".join(constructors)


def generate_constructor(form, classname, space_tag, coefficient_tag=None):
    "Generate a single Form constructor according to the given parameters."

    # Extract correct code snippets
    (argument, assign) = snippets[space_tag]

    # Construct list of arguments and function space assignments
    name = "V%d"
    if form.rank > 0:
        arguments = [argument % (name % i) for i in reversed(range(form.rank))]
        assignments = [assign % (i, name % i) for i in range(form.rank)]
    else:
        arguments = [argument]
        assignments = [assign]

    # Add coefficients to argument/assignment lists if specified
    if coefficient_tag is not None:
        (argument, assign) = snippets[coefficient_tag]
        arguments += [argument % name for name in form.coefficient_names]
        if form.rank > 0:  # FIXME: To match old generated code only
            assignments += [""]
        assignments += [assign % (name, name) for name in form.coefficient_names]

    # Add assignment of _ufc_form variable
    line = "\n    _ufc_form = std::make_shared<const %s>();"
    # FIXME: To match old generated code only
    if form.rank == 0 and coefficient_tag is None:
        line = "    _ufc_form = std::make_shared<const %s>();"
    assignments += [line % form.ufc_form_classname]

    # Construct list for initialization of Coefficient references
    initializers = ["%s(*this, %d)" % (name, number)
                    for (number, name) in enumerate(form.coefficient_names)]

    # Join lists together
    arguments = ", ".join(arguments)
    initializers = ", " + ", ".join(initializers) if initializers else ""
    body = "\n".join(assignments)

    # Wrap code into template
    args = {"classname": classname,
            "rank": form.rank,
            "num_coefficients": form.num_coefficients,
            "arguments": arguments,
            "initializers": initializers,
            "body": body,
            "superclass": form.superclassname
            }
    code = form_constructor_template % args
    return code


def generate_multimesh_constructor(form, classname, space_tag,
                                   coefficient_tag=None):
    "Generate a single Form constructor according to the given parameters."

    # Extract correct code snippets
    (argument, dummy) = snippets[space_tag]

    # Construct list of arguments
    name = "V%d"
    if form.rank > 0:
        arguments = [argument % (name % i) for i in reversed(range(form.rank))]
        spaces = [name % i for i in reversed(range(form.rank))]
    else:
        arguments = [argument]
        spaces = "mesh"

    # Add coefficients to argument/assignment lists if specified
    assignments = []
    if coefficient_tag is not None:
        (argument, assign) = snippets[coefficient_tag]
        arguments += [argument % name for name in form.coefficient_names]
        if form.rank > 0:  # FIXME: To match old generated code only
            assignments += [""]
        assignments += [assign % (name, name) for name in form.coefficient_names]

    # Construct list for initialization of Coefficient references
    initializers = ["%s(*this, %d)" % (name, number)
                    for (number, name) in enumerate(form.coefficient_names)]

    # Join lists together
    arguments = ", ".join(arguments)
    initializers = ", " + ", ".join(initializers) if initializers else ""

    # Ignore if functional
    if form.rank != 0:
        spaces = ", ".join(spaces)

    # Set access method
    if space_tag == "multimesh_shared_ptr_space":
        access = "->"
    else:
        access = "."

    # Create body for building multimesh function space
    body = ""
    body += "    // Create and add standard forms\n"
    body += "    std::size_t num_parts = V0%snum_parts(); // assume all equal and pick first\n" % access
    body += "    for (std::size_t part = 0; part < num_parts; part++)\n"
    body += "    {\n"
    body += "      std::shared_ptr<dolfin::Form> a(new %s(%s));\n" % (classname, ", ".join("V%d%spart(part)" % (i, access) for i in reversed(range(form.rank))))
    body += "    add(a);\n\n"
    body += "    }\n"
    body += "    // Build multimesh form\n"
    body += "    build();\n"

    # FIXME: Issue #91
    if form.rank == 0:
        body = ""
        body += "    // Creating a form for each part of the mesh\n"
        body += "    for (std::size_t i=0; i< mesh->num_parts(); i++)\n"
        body += "    {\n"
        body += "      std::shared_ptr<dolfin::Form> a(new %s(mesh->part(i))); " % (classname)
        body += "      add(a);"
        body += "    }\n"
        body += "    // Build multimesh form\n"
        body += "    build();\n"


    # Create body for assigning coefficients
    body += "\n    /// Assign coefficients"
    body += "\n".join(assignments) + "\n"

    # Wrap code into template
    args = {"classname": classname,
            "rank": form.rank,
            "num_coefficients": form.num_coefficients,
            "arguments": arguments,
            "spaces": spaces,
            "initializers": initializers,
            "body": body,
            "superclass": form.superclassname
            }
    code = multimesh_form_constructor_template % args
    return code


form_class_template = """\
class %(classname)s: public dolfin::%(superclass)s
{
public:

%(constructors)s

  // Destructor
  ~%(classname)s()
  {}

  /// Return the number of the coefficient with this name
  virtual std::size_t coefficient_number(const std::string& name) const
  {
%(coefficient_number)s
  }

  /// Return the name of the coefficient with this number
  virtual std::string coefficient_name(std::size_t i) const
  {
%(coefficient_name)s
  }

%(members)s
};
"""

multimesh_form_class_template = """\
class MultiMesh%(classname)s: public dolfin::MultiMesh%(superclass)s
{
public:

%(constructors)s

  // Destructor
  ~MultiMesh%(classname)s()
  {}

  /// Return the number of the coefficient with this name
  virtual std::size_t coefficient_number(const std::string& name) const
  {
%(coefficient_number)s
  }

  /// Return the name of the coefficient with this number
  virtual std::string coefficient_name(std::size_t i) const
  {
%(coefficient_name)s
  }

%(members)s
};
"""


# Template code for Form constructor
form_constructor_template = """\
  // Constructor
  %(classname)s(%(arguments)s):
    dolfin::%(superclass)s(%(rank)d, %(num_coefficients)d)%(initializers)s
  {
%(body)s
  }"""

multimesh_form_constructor_template = """\
  // Constructor
  MultiMesh%(classname)s(%(arguments)s):
    dolfin::MultiMesh%(superclass)s(%(spaces)s)%(initializers)s
  {
%(body)s
  }"""


def apply_form_template(classname, constructors, number, name, members,
                        superclass):
    args = {"classname": classname,
            "superclass": superclass,
            "constructors": constructors,
            "coefficient_number": number,
            "coefficient_name": name,
            "members": members}
    return form_class_template % args


def apply_multimesh_form_template(classname, constructors, number, name,
                                  members, superclass):
    members = members.replace("CoefficientAssigner",
                              "MultiMeshCoefficientAssigner")  # hack
    args = {"classname": classname,
            "superclass": superclass,
            "constructors": constructors,
            "coefficient_number": number,
            "coefficient_name": name,
            "members": members}
    return multimesh_form_class_template % args



# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import sys
import subprocess
import string

from setuptools import setup, find_packages

if sys.version_info < (2, 7):
    print("Python 2.7 or higher required, please upgrade.")
    sys.exit(1)

on_rtd = os.environ.get('READTHEDOCS') == 'True'

VERSION = "2017.1.0.post2"
RESTRICT_REQUIREMENTS = ">=2017.1,<2017.2"

if on_rtd:
    REQUIREMENTS = []
else:
    REQUIREMENTS = [
        "numpy",
        "six",
        "fenics-fiat%s" % RESTRICT_REQUIREMENTS,
        "fenics-ufl%s" % RESTRICT_REQUIREMENTS,
        "fenics-dijitso%s" % RESTRICT_REQUIREMENTS,
    ]

URL = "https://bitbucket.org/fenics-project/ffc/"

if sys.version_info[0] == 2:
    ENTRY_POINTS = {'console_scripts': ['ffc = ffc.__main__:main',
                                        'ffc-2 = ffc.__main__:main']}
else:
    ENTRY_POINTS = {'console_scripts': ['ffc = ffc.__main__:main',
                                        'ffc-3 = ffc.__main__:main']}

AUTHORS = """\
Anders Logg, Kristian Oelgaard, Marie Rognes, Garth N. Wells,
Martin Sandve Alnæs, Hans Petter Langtangen, Kent-Andre Mardal,
Ola Skavhaug, et al.
"""

CLASSIFIERS = """\
Development Status :: 5 - Production/Stable
Intended Audience :: Developers
Intended Audience :: Science/Research
License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)
Operating System :: POSIX
Operating System :: POSIX :: Linux
Operating System :: MacOS :: MacOS X
Operating System :: Microsoft :: Windows
Programming Language :: Python
Programming Language :: Python :: 2
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Topic :: Scientific/Engineering :: Mathematics
Topic :: Software Development :: Libraries :: Python Modules
Topic :: Software Development :: Code Generators
"""


def tarball():
    if "dev" in VERSION:
        return None
    return URL + "downloads/fenics-ffc-%s.tar.gz" % VERSION


def get_installation_prefix():
    "Get installation prefix"
    prefix = sys.prefix
    for arg in sys.argv[1:]:
        if "--user" in arg:
            import site
            prefix = site.getuserbase()
            break
        elif arg in ("--prefix", "--home", "--install-base"):
            prefix = sys.argv[sys.argv.index(arg) + 1]
            break
        elif "--prefix=" in arg or "--home=" in arg or "--install-base=" in arg:
            prefix = arg.split("=")[1]
            break

    return os.path.abspath(os.path.expanduser(prefix))


def get_git_commit_hash():
    """Return git commit hash of currently checked out revision
    or "unknown"
    """
    try:
        hash = subprocess.check_output(['git', 'rev-parse', 'HEAD'])
    except (OSError, subprocess.CalledProcessError) as e:
        print('Retrieving git commit hash did not succeed with exception:')
        print('"%s"' % str(e))
        print()
        print('Stored git commit hash will be set to "unknown"!')
        return "unknown"
    else:
        return hash.strip()


def write_config_file(infile, outfile, variables={}):
    "Write config file based on template"
    class AtTemplate(string.Template):
        delimiter = "@"
    s = AtTemplate(open(infile, "r").read())
    s = s.substitute(**variables)
    with open(outfile, "w") as a:
        a.write(s)


def generate_git_hash_file(GIT_COMMIT_HASH):
    "Generate module with git hash"
    write_config_file(os.path.join("ffc", "git_commit_hash.py.in"),
                      os.path.join("ffc", "git_commit_hash.py"),
                      variables=dict(GIT_COMMIT_HASH=GIT_COMMIT_HASH))


def run_install():
    "Run installation"

    # Get common variables
    #INSTALL_PREFIX = get_installation_prefix()
    GIT_COMMIT_HASH = get_git_commit_hash()

    # Scripts list
    entry_points = ENTRY_POINTS

    # Generate module with git hash from template
    generate_git_hash_file(GIT_COMMIT_HASH)

    # FFC data files
    data_files = [(os.path.join("share", "man", "man1"),
                  [os.path.join("doc", "man", "man1", "ffc.1.gz")])]

    # Call distutils to perform installation
    setup(name="fenics-ffc",
          description="The FEniCS Form Compiler",
          version=VERSION,
          author=AUTHORS,
          classifiers=[_f for _f in CLASSIFIERS.split('\n') if _f],
          license="LGPL version 3 or later",
          author_email="fenics-dev@googlegroups.com",
          maintainer_email="fenics-dev@googlegroups.com",
          url=URL,
          download_url=tarball(),
          platforms=["Windows", "Linux", "Solaris", "Mac OS-X", "Unix"],
          packages=find_packages("."),
          package_dir={"ffc": "ffc"},
          package_data={"ffc" : [os.path.join('backends', 'ufc', '*.h')]},
          #scripts=scripts,  # Using entry_points instead
          entry_points=entry_points,
          data_files=data_files,
          install_requires=REQUIREMENTS,
          zip_safe=False)


if __name__ == "__main__":
    run_install()



# -*- coding: utf-8 -*-
# Copyright (C) 2011-2016 Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.

from six import string_types
import numbers
import re

_subs = (
    # Remove 0s after e+ or e-
    (re.compile(r"e[\+]0*(.)"), r"e\1"),
    (re.compile(r"e[\-]0*(.)"), r"e-\1"),
    )
def format_float(x, precision=None):
    "Format a float value according to given precision."
    global _subs
    if precision:
        s = "{:.{prec}}".format(float(x), prec=precision)
    else:
        # Using "{}".format(float(x)) apparently results
        # in lesser precision in python 2 than python 3
        s = repr(float(x))
    for r, v in _subs:
        s = r.sub(v, s)
    return s


def format_int(x, precision=None):
    return str(x)


def format_value(value, precision=None):
    """Format a literal value as s tring.

    - float: Formatted according to current precision configuration.

    - int: Formatted as regular base 10 int literal.

    - str: Wrapped in "quotes".

    """
    if isinstance(value, numbers.Real):
        return format_float(float(value), precision=precision)
    elif isinstance(value, numbers.Integral):
        return format_int(int(value))
    elif isinstance(value, string_types):
        # FIXME: Is this ever used?
        assert '"' not in value
        return '"' + value + '"'
    elif hasattr(value, "ce_format"):
        return value.ce_format()
    else:
        raise RuntimeError("Unexpected type %s:\n%s" % (type(value), str(value)))
